from multiprocessing import Pipe
from queue import Queue
from threading import Thread

from mlgame.core.env import WS_WAIT_GAME_TIMEOUT
from mlgame.core.model import MLGameEntityWrapperSchema, MLGameDataType
from mlgame.utils.logger import logger


class CommunicationSet:
    """
    A data class for storing a set of communication objects and
    providing an interface for accessing these objects

    Communication objects used for receiving objects must provide `recv()` and `poll()`,
    and them used for sending objects must provide `send()`.
    For example, the object of `multiprocessing.connection.Connection` is a valid
    communication object for both sending and receiving.

    @var _recv_end A dictionary storing communication objects which are used to
         receive objects
    @var _send_end A dictionary storing communication objects which are used to
         send objects
    """

    def __init__(self):
        self._recv_end = {}
        self._send_end = {}

    def add_recv_end(self, name: str, comm_obj):
        """
        Add a new communication object for receiving objects

        @param name The name for distinguishing the added object. It could be used to
               distinguish the communication target.
        @param comm_obj The communication object which has `recv()` and `poll()` functions
        """
        if self._recv_end.get(name):
            raise ValueError("The name '{}' already exists in 'recv_end'".format(name))

        if not hasattr(comm_obj, "recv") or not hasattr(comm_obj, "poll"):
            raise ValueError("'comm_obj' doesn't have 'recv' or 'poll' function")

        self._recv_end[name] = comm_obj

    def add_send_end(self, name: str, comm_obj):
        """
        Add a new communication object for sending objects

        @param name The name for distinguishing the added object. It could be used to
               distinguish the communication target.
        @param comm_obj The communication object which has `send()` function
        """
        if self._send_end.get(name):
            raise ValueError("The name '{}' already exists in 'send_end'".format(name))

        if not hasattr(comm_obj, "send"):
            raise ValueError("'comm_obj' doesn't have 'send' function")

        self._send_end[name] = comm_obj

    def get_recv_end_names(self):
        """
        Get the name of all registered communication objects used for receiving

        @return A dictionary view objects containing the key of `_recv_end`
        """
        return self._recv_end.keys()

    def get_send_end_names(self):
        """
        Get the name of all registered communication objects used for sending

        @return A dictionary view objects containing the key of `_send_end`
        """
        return self._send_end.keys()

    def poll(self, name: str):
        """
        Check whether the specified communication object has data to read

        @param name The name of the communication object
        """
        return self._recv_end[name].poll()

    def recv(self, name: str, to_wait: bool = False):
        """
        Receive object from the specified communication object

        @param name The name of the communication object
        @param to_wait Whether to wait until the object is arrived
        @return The received object. If `to_wait` is False and nothing available from
                the specified communication object, return None.
        """
        if not to_wait and not self.poll(name):
            return None

        return self._recv_end[name].recv()

    def recv_all(self, to_wait: bool = False):
        """
        Receive objects from all communication object registered for receiving

        If `to_wait` is True, it will wait for the object one by one.

        @param to_wait Whether to wait until the object is arrived
        @return A dictionary storing received objects. The key is the name of
                the communication object, the value is the received object.
                If `to_wait` is False and nothing to receive, the value will be None.
        """
        objs = {}
        for comm_name in self._recv_end.keys():
            objs[comm_name] = self.recv(comm_name, to_wait)

        return objs

    def send(self, obj, name: str):
        """
        Send object via the specified communication object

        @param obj The object to be sent
        @param name The name of the communication object
        """
        self._send_end[name].send(obj)

    def send_all(self, obj):
        """
        Send object via all communication objects registered for sending

        @param obj The object to be sent
        """
        for comm_obj in self._send_end.values():
            comm_obj.send(obj)


class CommunicationHandler:
    """
    A data class for storing a sending and a receiving communication objects
    and providing interface for accessing them
    """

    def __init__(self):
        self._recv_end = None
        self._send_end = None

    def set_recv_end(self, comm_obj:Pipe):
        """
        Set the communication object for receiving

        @param comm_obj The communication object which has `recv` and `poll` function
        """
        if not hasattr(comm_obj, "recv") or not hasattr(comm_obj, "poll"):
            raise ValueError("'comm_obj' doesn't have 'recv' or 'poll' function")

        self._recv_end = comm_obj

    def set_send_end(self, comm_obj):
        """
        Set the communication object for sending

        @param comm_obj The communication object which has `send` function
        """
        if not hasattr(comm_obj, "send"):
            raise ValueError("'comm_obj' doesn't have 'send' function")

        self._send_end = comm_obj

    def poll(self):
        return self._recv_end.poll()

    def recv(self):
        return self._recv_end.recv()

    def send(self, obj):
        self._send_end.send(obj)


class GameCommManager:
    """
    The commnuication manager for the game process
    """

    def __init__(self):
        self._comm_to_ml_set = CommunicationSet()
        self._comm_to_others = CommunicationSet()

    def add_comm_to_ml(self, ml_name, recv_end, send_end):
        """
        Set communication objects for communicating with specified ml process
        """
        self._comm_to_ml_set.add_recv_end(ml_name, recv_end)
        self._comm_to_ml_set.add_send_end(ml_name, send_end)

    def get_ml_names(self):
        """
        Get the name of all registered ml process
        """
        return self._comm_to_ml_set.get_recv_end_names()

    def send_to_ml(self, obj, ml_name):
        """
        Send the object to the specified ml process
        """
        self._comm_to_ml_set.send(obj, ml_name)

    def send_to_all_ml(self, obj):
        """
        Send the object to all ml process
        """
        self._comm_to_ml_set.send_all(obj)

    def recv_from_ml(self, ml_name):
        """
        Receive the object from the specified ml process

        If the received object is `MLProcessError`, raise the exception.
        """
        obj = None
        while self._comm_to_ml_set.poll(ml_name):  # 有資料才讀
            obj = self._comm_to_ml_set.recv(ml_name,to_wait=False)

        return obj

    def recv_from_all_ml(self):
        """
        Receive objects from all the ml processes
        """
        obj_dict = {}
        for ml_name in self.get_ml_names():
            obj_dict[ml_name] = self.recv_from_ml(ml_name)
        return obj_dict

    def add_comm_to_others(self, client_name, recv_end, send_end):
        """
        Set communication objects for communicating with specified ml process
        """
        self._comm_to_others.add_recv_end(client_name, recv_end)
        self._comm_to_others.add_send_end(client_name, send_end)

    def send_to_others(self, obj:MLGameEntityWrapperSchema):
        """
        Send the object to all ml process
        """
        # print(obj)
        self._comm_to_others.send_all(obj)

    def recv_from_others(self):
        """
        Receive objects from all the ml processes
        """
        obj_dict = {}
        for client_name in self._comm_to_others.get_recv_end_names():
            obj_dict[client_name] = self._comm_to_others.recv(client_name, to_wait=False)
        return obj_dict


class MLCommManager:
    """
    The communication manager for the ml process
    """

    def __init__(self, ml_name):
        self._comm_to_game = CommunicationHandler()
        self._ml_name = ml_name

    def set_comm_to_game(self, recv_end:Pipe, send_end:Pipe):
        """
        Set communication objects for communicating with game process

        @param recv_end The communication object for receiving objects from game process
        @param send_end The communication object for sending objects to game process
        """
        self._comm_to_game.set_recv_end(recv_end)
        self._comm_to_game.set_send_end(send_end)

    def start_recv_obj_thread(self):
        """
        Start a thread to keep receiving objects from the game
        """
        self._obj_queue = Queue(2)

        thread = Thread(target=self._keep_recv_obj_from_game)
        thread.start()

    def _keep_recv_obj_from_game(self):
        """
        Keep receiving object from the game and put it in the queue

        If the queue is full, the received object will be dropped.
        """
        logger.info("ml listen")
        while True:
            if self._obj_queue.full():
                self._obj_queue.get_nowait()
                logger.warning(f"queue {self._ml_name} is full")
            while self._comm_to_game.poll():
                obj = self._comm_to_game.recv()
            self._obj_queue.put(obj)

            if obj is None:  # Received `None` from the game, quit the loop.
                break

    def recv_from_game(self) -> (dict, dict,bool):
        """
        Receive the most recent object from the game process.
        Discard any older objects in the queue.
        """
        try:
            data = ({},{},False)  # 假資料
            while self._comm_to_game.poll():  # 讀最新的資料
                data = self._comm_to_game.recv()
        except (EOFError, BrokenPipeError):
            logger.warning("AI pipe connection broken.")
            data = None
        return data
        # try:
        #     return self._obj_queue.get()
        # except Exception as e:
        #     logger.exception(e)
        #     return None

    def send_to_game(self, obj):
        """
        Send an object to the game process

        @param obj An object to be sent
        """
        try:
            self._comm_to_game.send(obj)
        except BrokenPipeError as e:
            print("Process '{}': The connection to the game process is closed."
                  .format(self._ml_name))
            logger.exception(e)


class TransitionCommManager:
    """
    The communication manager for the transition process
    """

    def __init__(self, recv_end=None, send_end=None,label=None):
        self._comm_to_game = CommunicationHandler()
        self.set_comm_to_game(recv_end, send_end)
        self.count = 0
        self.label = label

    def start_recv_obj_thread(self):
        """
        Start a thread to keep receiving objects from the game
        """
        self._obj_queue = Queue(1500)

        thread = Thread(target=self._keep_recv_obj_from_game)
        thread.start()

    def _keep_recv_obj_from_game(self):
        """
        Keep receiving object from the game and put it in the queue

        """
        while True:
            if self._obj_queue.full():
                self._obj_queue.get_nowait()
                logger.warning(f"The object queue for the process '{self.label}' is full. ")
            obj = self._comm_to_game.recv()
            logger.debug(f"obj:{obj}")
            self._obj_queue.put(obj)
            if obj is None:  # Received `None` from the game, quit the loop.
                break

    def set_comm_to_game(self, recv_end, send_end):
        """
        Set communication objects for communicating with game process

        @param recv_end The communication object for receiving objects from game process
        @param send_end The communication object for sending objects to game process
        """
        self._comm_to_game.set_recv_end(recv_end)
        self._comm_to_game.set_send_end(send_end)

    def recv_from_game(self)->MLGameEntityWrapperSchema:
        """
        Receive the object sent from the game process
        """
        try:
            obj = self._obj_queue.get(block=True, timeout=WS_WAIT_GAME_TIMEOUT)
            logger.debug(f"recv from game obj:{obj}")
            if obj:
                if isinstance(obj, MLGameEntityWrapperSchema):
                    return obj
                else:
                    return MLGameEntityWrapperSchema.model_validate(obj)
            return MLGameEntityWrapperSchema(type=MLGameDataType.NONE,data={})
        except Exception as e:
            logger.error(f"exception in process {self.label}")
            logger.exception(e)
            return MLGameEntityWrapperSchema(type=MLGameDataType.NONE,data={})

    def send_exception(self, exception):
        """
        Send an exception to the game process
        """
        self._comm_to_game.send(exception)



import time
import traceback

import pandas as pd
import pygame

from mlgame.core.communication import GameCommManager
from mlgame.core.exceptions import ErrorEnum, GameError, GameProcessError, MLProcessError
from mlgame.core.model import GameErrorSchema, MLGameDataType, MLGameEntityWrapperSchema, SystemMsgSchema, \
    GameProgressSchema
from mlgame.executor.interface import ExecutorInterface
from mlgame.game.generic import quit_or_esc
from mlgame.game.paia_game import PaiaGame
from mlgame.utils.io import save_json
from mlgame.utils.logger import logger
from mlgame.view.view import PygameViewInterface


class GameExecutor(ExecutorInterface):
    def __init__(
            self,
            game: PaiaGame,
            game_comm: GameCommManager,
            game_view: PygameViewInterface,
            fps=30, one_shot_mode=False, no_display=False, output_folder=None):
        self._view_data = None
        self._last_pause_btn_clicked_time = 0
        self._pause_state = False
        self.no_display = no_display
        self.game_view = game_view
        self.frame_count = 0
        self.game_comm = game_comm
        self.game = game
        self._active_ml_names = []
        self._ml_delayed_frames = {}
        self._active_ml_names = list(self.game_comm.get_ml_names())
        self._dead_ml_names = []
        self._ml_execution_time = 1 / fps
        self._fps = fps
        self._output_folder = output_folder
        self._ml_delay_counter = {}
        for name in self._active_ml_names:
            self._ml_delay_counter[name] = 0

        # self._recorder = get_recorder(self._execution_cmd, self._ml_names)
        self._game_executor_frame_count = 0
        self.one_shot_mode = one_shot_mode
        self._proc_name = str(self.game)
        self._clock = pygame.time.Clock()
        self._last_m_key_pressed = time.time()



    def run(self):
        game = self.game
        game_view = self.game_view
        try:
            self._send_system_message("AI準備中")

            self._send_game_info(game.get_scene_init_data())

            self._wait_all_ml_ready()
            self._send_system_message("遊戲啟動")
            while not self._quit_or_esc():
                if game_view.is_paused():
                    # 這裡的寫法不太好，但是可以讓遊戲暫停時，可以調整畫面。因為game_view裡面有調整畫面的程式。
                    self._clock.tick(self._fps)
                    game_view.draw(self._view_data)
                    # time.sleep(0.05)
                    game_view.sound_controller.pause()
                    continue

                scene_info_dict = game.get_data_from_game_to_player()
                keyboard_info = game_view.get_keyboard_info()

                self._send_data_to_ai_clients(scene_info_dict, keyboard_info)
                self._clock.tick(self._fps)
                cmd_dict = self._receive_cmd_from_ai_clients()

                # self._recorder.record(scene_info_dict, cmd_dict)
                logger.debug(f"game frame is {game.frame_count}")
                logger.debug(f"update game state at {self._game_executor_frame_count}")
                result = game.update(cmd_dict)

                logger.debug(f"get_scene_progress_data at {self._game_executor_frame_count}")

                self._view_data = game.get_scene_progress_data()

                logger.debug(f"game_view.draw at {self._game_executor_frame_count}")
                game_view.draw(self._view_data)

                logger.debug(f"play_audio at {self._game_executor_frame_count}")

                self.game_view.play_audio(self._view_data)
                self._game_executor_frame_count += 1

                # save image
                if self._output_folder:
                    game_view.save_image(f"{self._output_folder}/{self._game_executor_frame_count:05d}.jpg")
                self._send_game_progress(self._view_data)

                # Do reset stuff

                if result == "RESET" or result == "QUIT":
                    logger.debug(f"reset or quit by {result} at {self._game_executor_frame_count}")
                    scene_info_dict = game.get_data_from_game_to_player()
                    # send to ml_clients and don't parse any command , while client reset ,
                    # self._wait_all_ml_ready() will works and not blocks the process
                    for ml_name in self._active_ml_names:
                        self.game_comm.send_to_ml((scene_info_dict[ml_name], [],True), ml_name)
                    # TODO check what happen when bigfile is saved
                    time.sleep(0.1)
                    game_result = game.get_game_result()

                    attachments = game_result['attachment']
                    print(pd.DataFrame(attachments).to_string())

                    if self.one_shot_mode or result == "QUIT":
                        self._send_system_message("遊戲結束")
                        self._send_game_result(game_result)
                        if self._output_folder:
                            save_json(self._output_folder, game_result)
                        self._send_system_message("關閉遊戲")
                        self._send_end_message()
                        time.sleep(1)

                        break
                    game.reset()
                    game_view.reset()

                    self._game_executor_frame_count = 0
                    for name in self._active_ml_names:
                        self._ml_delay_counter[name] = 0
                    logger.debug(f"_wait_all_ml_ready at {self._game_executor_frame_count}")

                    self._wait_all_ml_ready()
        except Exception as e:
            # handle unknown exception
            # send to es
            logger.exception(e)
            e = GameProcessError(self._proc_name, traceback.format_exc())
            self._send_game_error_with_obj(GameError(
                error_type=ErrorEnum.GAME_EXEC_ERROR,
                message=e.__str__(),
                frame=self._game_executor_frame_count,
            ))
            self._send_system_message("遊戲結束")
            game_result = game.get_game_result()
            self._send_game_result(game_result)
            if self._output_folder:
                save_json(self._output_folder, game_result)
            self._send_system_message("關閉遊戲")
            self._send_end_message()
            time.sleep(1)

        pass

    def _wait_all_ml_ready(self):
        """
        Wait until receiving "READY" commands from all ml processes
        """
        # Wait the ready command one by one
        logger.info("waiting for all ai client ready")
        # add a delay let ai_client send data into pipe
        # time.sleep(0.1)
        for ml_name in self._active_ml_names:
            recv = ""
            while recv != "READY":
                try:
                    recv = self.game_comm.recv_from_ml(ml_name)
                    if isinstance(recv, GameError):
                        # handle error when ai_client couldn't be ready state.
                        self._dead_ml_names.append(ml_name)
                        self._active_ml_names.remove(ml_name)
                        self._send_game_error_with_obj(recv)
                        break
                except Exception as e:
                    logger.exception(e)
                    self._dead_ml_names.append(ml_name)
                    self._active_ml_names.remove(ml_name)
                    # self._send_game_error(f"AI of {ml_name} has error at initial stage.")
                    ai_error = GameError(
                        error_type=ErrorEnum.AI_INIT_ERROR, frame=0,
                        message=f"AI of {ml_name} has error at initial stage. {e.__str__()}")

                    self._send_game_error_with_obj(ai_error)

                    break
    def _send_data_to_ai_clients(self,scene_info_dict, keyboard_info):
        logger.debug(f"send_data_to_ai_clients  at {self._game_executor_frame_count}")
        try:
            # TODO add schema
            for ml_name in self._active_ml_names:
                self.game_comm.send_to_ml((scene_info_dict[ml_name], keyboard_info,self.game.ai_enabled), ml_name)
        except KeyError:
            raise KeyError(
                "The game doesn't provide scene information "
                f"for the client '{ml_name}'")
        pass
    def _receive_cmd_from_ai_clients(self) -> dict:
        logger.debug(f"receive_cmd_from_ai_clients at {self._game_executor_frame_count}")
        response_dict = self.game_comm.recv_from_all_ml()

        # logger.info(f"check command of {response_dict} from ml at {self._frame_count}")
        cmd_dict = {}
        for ml_name in self._active_ml_names[:]:
            cmd_received = response_dict[ml_name]
            if isinstance(cmd_received, MLProcessError):
                # print(cmd_received.message)
                # handle error from ai clients
                logger.error(cmd_received)
                self._send_game_error_with_obj(GameError(
                    error_type=ErrorEnum.AI_EXEC_ERROR,
                    message=str(cmd_received),
                    frame=self._game_executor_frame_count
                ))
                self._dead_ml_names.append(ml_name)
                self._active_ml_names.remove(ml_name)
            elif isinstance(cmd_received, GameError):
                logger.error(cmd_received)
                self._send_game_error_with_obj(cmd_received)
                self._dead_ml_names.append(ml_name)
                self._active_ml_names.remove(ml_name)
            elif isinstance(cmd_received, dict):
                self._check_delay(ml_name, cmd_received["frame"])
                cmd_dict[ml_name] = cmd_received["command"]
            else:
                # logger.warning(f"cmd is {cmd_received} at {self._frame_count}")
                cmd_dict[ml_name] = None

        for ml_name in self._dead_ml_names:
            cmd_dict[ml_name] = None

        if len(self._active_ml_names) == 0:
            error = MLProcessError(
                self._proc_name,
                f"The process {self._proc_name} exit because all ml processes has exited.")
            game_error = GameError(
                error_type=ErrorEnum.GAME_EXEC_ERROR, frame=self._game_executor_frame_count,
                message="All ml clients has been terminated")

            self._send_game_error_with_obj(game_error)
            self._send_game_result(self.game.get_game_result())
            self._send_end_message()

            raise error
        return cmd_dict

        pass

    
    def _check_delay(self, ml_name, cmd_frame):
        """
        Check if the timestamp of the received command is delayed
        """
        delayed_frame = self.game.frame_count - cmd_frame

        if delayed_frame > 0:
            self._ml_delay_counter[ml_name]+=1
            # logger.warning(
            #     f"AI({ml_name}) 在第 {self.game.frame_count} 遊戲幀，延遲了 {delayed_frame} 幀，目前已經延遲 {self._ml_delay_counter[ml_name]} 次"
            # )
            logger.warning(
                f"AI({ml_name}) delay {delayed_frame} frame at game_frame({self.game.frame_count}), it has been accumulated {self._ml_delay_counter[ml_name]} times"
            )
            self._send_game_error_with_obj(
                GameError(
                    error_type=ErrorEnum.AI_EXEC_ERROR,
                    message=f"AI({ml_name}) delay {delayed_frame} frame at game_frame({self.game.frame_count}), it has been accumulated {self._ml_delay_counter[ml_name]} times",
                    frame=self._game_executor_frame_count
                )
            )


    def _quit_or_esc(self) -> bool:
        if self.no_display:
            return self._game_executor_frame_count > 30000
        else:
            return quit_or_esc()

    def _send_game_result(self, game_result_dict):
        # TO be deprecated("_send_game_error_with_obj")
        obj = MLGameEntityWrapperSchema(
                type=MLGameDataType.GAME_RESULT,data=game_result_dict)
        self.game_comm.send_to_others(
            obj
        )

    def _send_system_message(self, msg: str):

        self.game_comm.send_to_others(
            MLGameEntityWrapperSchema(
                type=MLGameDataType.SYSTEM_MSG,data=SystemMsgSchema(message= msg))
        )

    def _send_game_info(self, game_info_dict):

        self.game_comm.send_to_others(
            MLGameEntityWrapperSchema(
                type=MLGameDataType.GAME_INFO,data=game_info_dict
            )
        )

    def _send_game_progress(self, game_progress_dict:GameProgressSchema):
        """
        Send the game progress to the transition server
        """
        if not isinstance(game_progress_dict, GameProgressSchema):
            game_progress_dict = GameProgressSchema.model_validate(game_progress_dict)
        game_progress_dict.frame = self._game_executor_frame_count
        #
        # data_dict = {
        #     "type": "game_progress",
        #     "data": game_progress_dict
        # }


        # use wrapper
        # logger.debug(game_progress_dict)
        logger.debug(f"send game progress at {self._game_executor_frame_count}")
        self.game_comm.send_to_others(
            MLGameEntityWrapperSchema(
            data=game_progress_dict
        ))


    def _send_game_error_with_obj(self, error: GameError):
        # logger.error(error)

        self.game_comm.send_to_others( MLGameEntityWrapperSchema(
                type=MLGameDataType.GAME_ERROR,
                data=GameErrorSchema(
                    message=error.message,
                    error_type=error.error_type,
                    frame=error.frame),
            ))


    def _send_end_message(self):
        self.game_comm.send_to_others(MLGameEntityWrapperSchema(type=MLGameDataType.END))




import math

import pygame

from mlgame.game.paia_game import PaiaGame
from mlgame.utils.enum import get_ai_name
from mlgame.view.audio_model import MusicProgressSchema, create_sound_init_data, create_music_init_data
from mlgame.view.decorator import check_game_progress, check_game_result, check_scene_init_data
from mlgame.view.view_model import create_text_view_data, create_asset_init_data, create_image_view_data, \
    create_line_view_data, Scene, create_polygon_view_data, create_rect_view_data, create_scene_progress_data
from .env import *
from .env import HELP_TXT_STYLE, HELP_TXT_COLOR
from .mazeMode import MazeMode


class Dont_touch(PaiaGame):
    def __init__(self, user_num, map_num, time_to_play, dark_mode='dark', map_file=None, *args, **kwargs):
        super().__init__(user_num=user_num)
        # self.game_type = game_type
        sensor_num = 6
        self.user_num = user_num
        if dark_mode == 'dark':
            self.dark_mode = True
        elif dark_mode == 'light':
            self.dark_mode = False
        else:
            self.dark_mode = True
        self.is_single = False
        if self.user_num == 1:
            self.is_single = True
        # self.maze_id = map - 1
        if map_file is None:
            map_file = path.join(MAP_DIR, f"level_{map_num}.tmj")
        self.map_file = map_file
        self.game_end_time = time_to_play
        self.sensor_num = sensor_num
        self.set_game_mode()
        self.is_running = self.isRunning()
        self.map_width = self.game_mode.map.width
        self.map_height = self.game_mode.map.height
        self.scene = Scene(WIDTH, HEIGHT, "#08142b", 0, 0)
        self.origin_car_pos = [0, 0]

    # self.origin_car_pos = self.game_mode.car_info[0]["center"]

    def update(self, cmd_dict):
        self.frame_count += 1
        self.game_mode.handle_event()
        self.game_mode.update_sprite(cmd_dict)
        if not self.isRunning():
            self.is_running = False
            return "RESET"
        for car in self.game_mode.cars:
            if self.origin_car_pos != [0, 0]:
                break
            self.origin_car_pos = car.get_info()["center"]

    def get_data_from_game_to_player(self):
        scene_info = self.get_scene_info
        player_info = {}
        # pygame -> box2D
        end_p = self.game_mode.pygame_to_box2d(self.game_mode.end_point.get_info()["coordinate"], self.map_height / PPM)
        check_points_coodinate = []
        for cp in self.game_mode.check_points:
            check_points_coodinate.append(
                self.game_mode.pygame_to_box2d(cp.get_info()["coordinate"], self.map_height / PPM))
        for car in self.game_mode.car_info:
            # type of car is dictionary
            player_info[get_ai_name(int(car["id"]))] = {
                "frame": scene_info["frame"],
                "status": car["status"],
                "x": car["coordinate"][0],
                "y": car["coordinate"][1],
                "angle": (car["angle"] * 180 / math.pi) % 360,
                "R_sensor": car["r_sensor_value"]["distance"],
                "L_sensor": car["l_sensor_value"]["distance"],
                "F_sensor": car["f_sensor_value"]["distance"],
                "B_sensor": car["b_sensor_value"]["distance"],
                "L_T_sensor": car["l_t_sensor_value"]["distance"],
                "R_T_sensor": car["r_t_sensor_value"]["distance"],
                "crash_count": car["crash_times"],
                "end_x": end_p[0],
                "end_y": end_p[1],
                "check_points": check_points_coodinate
            }
        return player_info

    def reset(self):
        self.frame_count = 0
        self.set_game_mode()
        # self.game_mode.sound_controller.play_music()

    def isRunning(self):
        return self.game_mode.isRunning()

    @property
    def get_scene_info(self):
        """
        Get the scene information
        """
        scene_info = {
            "frame": self.game_mode.frame,
        }

        for car in self.game_mode.car_info:
            # type of car is dictionary
            scene_info[str(car["id"]) + "P_position"] = car["topleft"]
        return scene_info

    @check_scene_init_data
    def get_scene_init_data(self) -> dict:
        """
        Get the scene and object information for drawing on the web
        """
        game_info = {
            "scene": self.scene.__dict__, "assets": [], "background": [],
            "map_width": self.game_mode.map.tileWidth * 20, "map_height": self.game_mode.map.tileHeight * 20
        }

        game_info["assets"].append(create_asset_init_data("target", 40, 40, TARGET_PATH, TARGET_URL))
        if self.dark_mode:
            bg_path = path.join(ASSET_IMAGE_DIR, BG_IMG)
            bg_url = BG_URL
            game_info["assets"].append(create_asset_init_data("bg_img", 600, 600, bg_path, bg_url))
        else:
            bg_path = path.join(ASSET_IMAGE_DIR, L_BG_IMG)
            bg_url = L_BG_URL
            game_info["assets"].append(create_asset_init_data("bg_img", 600, 600, bg_path, bg_url))
        bar_path = path.join(ASSET_IMAGE_DIR, BAR_IMG)
        bar_url = BAR_URL
        game_info["assets"].append(create_asset_init_data("bar_img", 600, 600, bar_path, bar_url))
        for i in range(self.user_num):
            game_info["assets"].append(
                create_asset_init_data(f"car_0{i + 1}", 32, 32, path.join(ASSET_IMAGE_DIR, f"car_0{i + 1}.png"),
                                       f"https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/dont_touch/{version}/asset/image/car_0{i + 1}.png"))
        for i in range(0, 6):
            game_info["assets"].append(create_asset_init_data(f"regularExplosion0{i}", 40, 40,
                                                              path.join(ASSET_IMAGE_DIR, f"regularExplosion0{i}.png"),
                                                              f"https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/dont_touch/{version}/asset/image/regularExplosion0{i}.png"))

        game_info["background"].append(create_image_view_data("bg_img", 0, 0, 640, 640))
        game_info["background"].append(create_image_view_data("bar_img", 640, 0, 160, 640))

        # game_info["background"].append(create_image_view_data("target", 0, 0, 640, 640))

        for wall in self.game_mode.walls:
            vertices = [(wall.body.transform * v) for v in wall.box.shape.vertices]
            vertices = [self.game_mode.transfer_box2d_to_pygame(v) for v in vertices]
            game_info["background"].append(create_polygon_view_data("wall", vertices, PAIA_B))
        for wall in self.game_mode.slant_walls:
            vertices = [(wall.body.transform * v) for v in wall.box.shape.vertices]
            vertices = [self.game_mode.transfer_box2d_to_pygame(v) for v in vertices]
            game_info["background"].append(create_polygon_view_data("wall", vertices, PAIA_B))
        game_info["musics"] = [
            create_music_init_data("bgm", file_path=BGM_PATH, github_raw_url=BGM_URL),

        ]
        game_info["sounds"] = [
            create_sound_init_data("bomb", file_path=BOMB_PATH, github_raw_url=BOMB_URL),

        ]
        return game_info

    @check_game_progress
    def get_scene_progress_data(self) -> dict:
        """
        Get the position of game objects for drawing on the web
        """
        background = []
        object_list = []
        toggle_with_bias = []
        toggle = []
        game_sys_info = {}
        sound = self.game_mode.sound

        game_progress = {
            "frame": self.frame_count,
            "background": [],
            "object_list": [],
            "toggle_with_bias": [],
            "toggle": [],
            "foreground": [],
            "user_info": [],
            "game_sys_info": {}
        }
        game_sys_info = {"view_center_coordinate": [0, 0]}
        for p in self.game_mode.all_points:
            object_list.append(p.get_progress_data())
            # if isinstance(p,End_point):
            #     point_info = p.get_info()
            x, y = self.game_mode.pygame_to_box2d(p.get_info()["coordinate"], self.map_height / PPM)
            toggle_with_bias.append(
                create_rect_view_data("car_coordinate_rect",
                                      p.rect.centerx - 40, p.rect.centery, 90, 20, BLACK))

            toggle_with_bias.append(
                create_text_view_data(
                    f"({x:.2f},{y:.2f})",
                    p.rect.centerx - 40, p.rect.centery,
                    HELP_TXT_COLOR,
                    HELP_TXT_STYLE
                )
            )

        # background.append(create_image_view_data("bar_img", 640, 0, 160, 640))
        # wall

        # text
        background.append(
            create_text_view_data("{0:05d} frames".format(self.frame_count), 658, 30, WHITE, font_style="21px Arial"))
        for car in self.game_mode.car_info:
            background.append(
                create_text_view_data(
                    f"{car['crash_times']}", WIDTH - 108, 104 + 140 * car["id"], WHITE, font_style="20px Arial")
            )

            background.append(
                create_text_view_data(
                    f"{car['check_point']}", WIDTH - 108, 142 + 140 * car["id"], WHITE,
                    font_style="20px Arial"))

            background.append(
                create_text_view_data(
                    "{0:04d} frames".format(car["end_frame"]), WIDTH - 114, 179 + 140 * car["id"], WHITE,
                    font_style="20px Arial"))

            if car["is_running"]:
                # line
                object_list.append(
                    create_line_view_data(
                        "l_sensor", car["center"][0], car["center"][1],
                        self.game_mode.transfer_box2d_to_pygame(car["l_sensor_value"]["coordinate"])[
                            0],
                        self.game_mode.transfer_box2d_to_pygame(car["l_sensor_value"]["coordinate"])[
                            1],
                        SENSOR_R, 5))

                object_list.append(
                    create_line_view_data("l_top_sensor", car["center"][0], car["center"][1],
                                          self.game_mode.transfer_box2d_to_pygame(
                                              car["l_t_sensor_value"]["coordinate"])[
                                              0],
                                          self.game_mode.transfer_box2d_to_pygame(
                                              car["l_t_sensor_value"]["coordinate"])[
                                              1],
                                          SENSOR_R, 5))

                object_list.append(
                    create_line_view_data("r_top_sensor", car["center"][0], car["center"][1],
                                          self.game_mode.transfer_box2d_to_pygame(
                                              car["r_t_sensor_value"]["coordinate"])[
                                              0],
                                          self.game_mode.transfer_box2d_to_pygame(
                                              car["r_t_sensor_value"]["coordinate"])[
                                              1],
                                          SENSOR_B, 5))
                object_list.append(
                    create_line_view_data("r_sensor", car["center"][0], car["center"][1],
                                          self.game_mode.transfer_box2d_to_pygame(car["r_sensor_value"]["coordinate"])[
                                              0],
                                          self.game_mode.transfer_box2d_to_pygame(car["r_sensor_value"]["coordinate"])[
                                              1],
                                          SENSOR_B, 5))
                object_list.append(
                    create_line_view_data("f_sensor", car["center"][0], car["center"][1],
                                          self.game_mode.transfer_box2d_to_pygame(car["f_sensor_value"]["coordinate"])[
                                              0],
                                          self.game_mode.transfer_box2d_to_pygame(car["f_sensor_value"]["coordinate"])[
                                              1],
                                          SENSOR_Y, 5))
                object_list.append(
                    create_line_view_data("b_sensor", car["center"][0], car["center"][1],
                                          self.game_mode.transfer_box2d_to_pygame(car["b_sensor_value"]["coordinate"])[
                                              0],
                                          self.game_mode.transfer_box2d_to_pygame(car["b_sensor_value"]["coordinate"])[
                                              1],
                                          SENSOR_Y, 5))
                # sensor value
                toggle_with_bias.append(
                    create_rect_view_data("sensor_rect", 60 * math.sin(car["angle"]) + car["center"][0] - 25,
                                          60 * math.cos(car["angle"]) + car["center"][1] - 10, 50, 20, BLACK))  # behind
                toggle_with_bias.append(
                    create_rect_view_data("sensor_rect", 60 * math.sin(-car["angle"]) + car["center"][0] - 25,
                                          -60 * math.cos(-car["angle"]) + car["center"][1] - 10, 50, 20,
                                          BLACK))  # front
                toggle_with_bias.append(create_rect_view_data("sensor_rect", 60 * math.sin(
                    -car["angle"] + math.pi / 4) + car["center"][0] - 25, -60 * math.cos(-car["angle"] + math.pi / 4) +
                                                              car["center"][1] - 10, 50, 20, BLACK))
                toggle_with_bias.append(create_rect_view_data("sensor_rect", 60 * math.sin(
                    -car["angle"] + math.pi / 2) + car["center"][0] - 25, -60 * math.cos(-car["angle"] + math.pi / 2) +
                                                              car["center"][1] - 10, 50, 20, BLACK))
                toggle_with_bias.append(create_rect_view_data("sensor_rect", 60 * math.sin(
                    -car["angle"] - math.pi / 4) + car["center"][0] - 25, -60 * math.cos(-car["angle"] - math.pi / 4) +
                                                              car["center"][1] - 10, 50, 20, BLACK))
                toggle_with_bias.append(create_rect_view_data("sensor_rect", 60 * math.sin(
                    -car["angle"] - math.pi / 2) + car["center"][0] - 25, -60 * math.cos(-car["angle"] - math.pi / 2) +
                                                              car["center"][1] - 10, 50, 20, BLACK))

                toggle_with_bias.append(create_text_view_data(f"{car['b_sensor_value']['distance']}",
                                                              60 * math.sin(car["angle"]) +
                                                              car["center"][0] - 20,
                                                              60 * math.cos(car["angle"]) +
                                                              car["center"][1] - 10, SENSOR_Y,
                                                              font_style="20px Arial"))
                toggle_with_bias.append(create_text_view_data(f"{car['f_sensor_value']['distance']}",
                                                              60 * math.sin(-car["angle"]) +
                                                              car["center"][0] - 20,
                                                              -60 * math.cos(-car["angle"]) +
                                                              car["center"][1] - 10, SENSOR_Y,
                                                              font_style="20px Arial"))
                toggle_with_bias.append(create_text_view_data(f"{car['r_t_sensor_value']['distance']}",
                                                              60 * math.sin(
                                                                  -car["angle"] + math.pi / 4) +
                                                              car["center"][0] - 20, -60 * math.cos(
                        -car["angle"] + math.pi / 4) + car["center"][1] - 10, SENSOR_B, font_style="20px Arial"))
                toggle_with_bias.append(create_text_view_data(f"{car['r_sensor_value']['distance']}",
                                                              60 * math.sin(
                                                                  -car["angle"] + math.pi / 2) +
                                                              car["center"][0] - 20, -60 * math.cos(
                        -car["angle"] + math.pi / 2) + car["center"][1] - 10, SENSOR_B, font_style="20px Arial"))
                toggle_with_bias.append(create_text_view_data(f"{car['l_t_sensor_value']['distance']}",
                                                              60 * math.sin(
                                                                  -car["angle"] - math.pi / 4) +
                                                              car["center"][0] - 20, -60 * math.cos(
                        -car["angle"] - math.pi / 4) + car["center"][1] - 10, SENSOR_R, font_style="20px Arial"))
                toggle_with_bias.append(create_text_view_data(f"{car['l_sensor_value']['distance']}",
                                                              60 * math.sin(
                                                                  -car["angle"] - math.pi / 2) +
                                                              car["center"][0] - 20, -60 * math.cos(
                        -car["angle"] - math.pi / 2) + car["center"][1] - 10, SENSOR_R, font_style="20px Arial"))

                # 車子座標
                toggle_with_bias.append(
                    create_rect_view_data("car_coordinate_rect",
                                          car["topleft"][0] - 20, car["center"][1] + 5, 90, 20, BLACK))

                toggle_with_bias.append(
                    create_text_view_data(
                        f"({car['coordinate'][0]:.1f},{car['coordinate'][1]:.1f})",
                        car["topleft"][0] - 20, car["center"][1] + 5,
                        HELP_TXT_COLOR,
                        HELP_TXT_STYLE
                    )
                )
            object_list.append(
                create_image_view_data(car["image"], car["topleft"][0], car["topleft"][1], car["size"][0],
                                       car["size"][1],
                                       car["angle"])
            )
        game_progress = create_scene_progress_data(
            frame=self.frame_count,
            background=background,
            object_list=object_list,
            foreground=[],
            toggle=toggle,
            toggle_with_bias=toggle_with_bias,
            musics=[MusicProgressSchema(music_id=f"bgm").__dict__],
            sounds=sound, game_sys_info=game_sys_info
        )
        return game_progress

    @check_game_result
    def get_game_result(self):
        """
        Get the game result for the web
        """
        scene_info = self.get_scene_info
        result = self.game_mode.result
        rank = []
        for user in self.game_mode.ranked_user:
            if self.game_mode.check_point_num:
                pass_percent = round(user.check_point / self.game_mode.check_point_num, 5) * 100
                remain_point = self.game_mode.check_point_num - user.check_point
            else:
                pass_percent = 0
                remain_point = 0
            same_rank = {"player_num": str(user.car_no + 1) + "P",
                         "rank": self.game_mode.ranked_user.index(user) + 1,
                         # "frame_limit": self.game_end_time,
                         "used_frame": user.end_frame,
                         # "frame_percent": round(user.end_frame / self.game_end_time * 100, 3),
                         "total_checkpoints": self.game_mode.check_point_num,
                         "check_points": user.check_point,
                         # "remain_points": remain_point,
                         # "pass_percent": pass_percent,
                         "crash_count": user.collide_times,
                         "score": 10000 * user.check_point - 0.001 * user.end_frame - 10 * user.collide_times
                         }
            rank.append(same_rank)
        print({"frame_used": scene_info["frame"],
               "status": self.game_mode.state,
               "attachment": rank,
               })
        return {"frame_used": scene_info["frame"],
                "status": self.game_mode.state,
                "attachment": rank,
                }

        pass

    def get_keyboard_command(self):
        """
        Get the command according to the pressed keys
        """
        if not self.isRunning():
            return {"1P": "RESET",
                    "2P": "RESET",
                    "3P": "RESET",
                    "4P": "RESET"
                    }
        key_pressed_list = pygame.key.get_pressed()
        cmd_1P = {"left_PWM": 0, "right_PWM": 0}
        cmd_2P = {"left_PWM": 0, "right_PWM": 0}
        cmd_3P = {"left_PWM": 0, "right_PWM": 0}
        cmd_4P = {"left_PWM": 0, "right_PWM": 0}

        if key_pressed_list[pygame.K_UP]:
            cmd_1P["left_PWM"] = 100
            cmd_1P["right_PWM"] = 100
        elif key_pressed_list[pygame.K_DOWN]:
            cmd_1P["left_PWM"] = -100
            cmd_1P["right_PWM"] = -100
        if key_pressed_list[pygame.K_LEFT]:
            cmd_1P["right_PWM"] += 100
        elif key_pressed_list[pygame.K_RIGHT]:
            cmd_1P["left_PWM"] += 100

        if key_pressed_list[pygame.K_w]:
            cmd_2P["left_PWM"] = 100
            cmd_2P["right_PWM"] = 100
        elif key_pressed_list[pygame.K_s]:
            cmd_2P["left_PWM"] = -100
            cmd_2P["right_PWM"] = -100
        elif key_pressed_list[pygame.K_a]:
            cmd_2P["right_PWM"] += 100
        elif key_pressed_list[pygame.K_d]:
            cmd_2P["left_PWM"] += 100

        return {"1P": cmd_1P,
                "2P": cmd_2P,
                "3P": cmd_3P,
                "4P": cmd_4P}

    def set_game_mode(self):

        self.game_mode = MazeMode(self.user_num, self.map_file, self.game_end_time, self.sensor_num)
        # self.game_type = "MAZE"



import pygame
import os
# Initialize pygame
def test_draw_ch_txt():
    pygame.init()

    # Screen setup
    WIDTH, HEIGHT = 600, 400
    screen = pygame.display.set_mode((WIDTH, HEIGHT))
    pygame.display.set_caption("Pygame Chinese Text Example")

    # Load a Chinese font (Ensure this font file exists)
    font_path = os.path.join(os.path.dirname(__file__),"..","assets", "NotoSansTC-Regular.ttf")  # Replace with your own Chinese TTF/OTF font
    font = pygame.font.Font(font_path, 36)  # Load font with size 36

    # Render text
    text_surface = font.render("你好，Pygame!", True, (0, 0, 0))  # Black color

    # Game loop
    running = True
    while running:
        screen.fill((255, 255, 255))  # White background
        screen.blit(text_surface, (100, 150))  # Draw text on screen

        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                running = False

        pygame.display.flip()

    pygame.quit()


"""
This is a base class for different mode in game.
"""
from .car import Car
from .points import *
from .maze_wall import SlantWall

import pygame

from .env import *


class GameMode(object):
    def __init__(self, bg_img=pygame.Surface((WIDTH, HEIGHT))):
        self.bg_img = bg_img
        self.clock = pygame.time.Clock()
        self.running = True
        self.frame = 0
        pygame.font.init()
        self.font = pygame.font.Font(pygame.font.match_font("arial", bold=True), 15)
        self.time_font = pygame.font.Font(pygame.font.match_font("arial", bold=True), 46)
        self.check_point_num = 0
        # self.start_time = time.time()
        self.ranked_user = []  # pygame.sprite car


    def ticks(self, fps=FPS):
        """This method should be called once per frame.
        It will compute how many milliseconds have passed since the previous call.
        :param fps: frame per second 每秒的繪圖次數
        :return: None
        """
        self.clock.tick(fps)

    def handle_event(self):
        """ Handle the event from window , mouse or button.
        :return: None
        """
        pass

    def detect_collision(self):
        """ Detect the collision event between sprites.
        :return: None
        """
        pass

    def update_sprite(self, *args):
        """ This function should update every sprite in games.
        :return: None
        """
        pass

    def draw_bg(self):
        """  Draw a background on screen.
        :return:None
        """
        pass

    def drawWorld(self):
        """  This function should draw every sprite on specific surface.
        :return: None
        """
        pass

    def flip(self):
        """Update the full display Surface to the screen
        :return:None
        """
        pygame.display.flip()

    def isRunning(self) -> bool:
        return self.running

    def rank(self)->list[Car]:

        sorted_cars = sorted(self.eliminated_user, key=lambda x: x.score, reverse=True)

        # Initialize rank and previous score
        rank = 1
        previous_score = None
        self.ranked_user = []
        # Iterate through sorted cars and assign ranks
        for i, car in enumerate(sorted_cars):
            if previous_score is None or car.score != previous_score:
                rank = i + 1
            car.rank = rank
            previous_score = car.score
            self.ranked_user.append(car)

        return self.ranked_user

    def trnsfer_box2d_to_pygame(self, coordinate):
        '''
        :param coordinate: vertice of body of box2d object
        :return: center of pygame rect
        '''
        return ((coordinate[0] - self.pygame_point[0]) * PPM, (self.pygame_point[1] - coordinate[1]) * PPM)

    def get_wall_info_v(self, wall_tile):
        wall_tiles = []
        for col in range(len(self.map.data[0]) - 1):
            row = 0
            first_tile = -1
            last_tile = -1
            while row < len(self.map.data):
                tiles = self.map.data[row]

                if (tiles[col]%18) == wall_tile:
                    if first_tile == -1:
                        first_tile = row
                        if row == len(self.map.data) - 1:
                            last_tile = row
                            self.wall_vertices_for_Box2D.append(
                                {"type": wall_tile,
                                 "vertices": self.wall_vertices_v((col, first_tile), (col, last_tile))})

                            # self.wall_vertices_for_Box2D.append(self.wall_vertices_v((col, first_tile), (col, last_tile)))
                            first_tile = -1
                            row += 1
                        else:
                            row += 1
                    elif row == len(self.map.data) - 1:
                        last_tile = row
                        self.wall_vertices_for_Box2D.append(
                            {"type": wall_tile, "vertices": self.wall_vertices_v((col, first_tile), (col, last_tile))})
                        # self.wall_vertices_for_Box2D.append(self.wall_vertices_v((col, first_tile), (col, last_tile)))
                        first_tile = -1
                        row += 1
                    else:
                        row += 1
                else:
                    if first_tile != -1:
                        last_tile = row - 1
                        self.wall_vertices_for_Box2D.append(
                            {"type": wall_tile, "vertices": self.wall_vertices_v((col, first_tile), (col, last_tile))})
                        # self.wall_vertices_for_Box2D.append(self.wall_vertices_v((col, first_tile), (col, last_tile)))
                        first_tile = -1
                        row += 1
                    else:
                        row += 1

    def get_wall_info_h(self, wall_tile):
        wall_tiles = []
        for row, tiles in enumerate(self.map.data):
            col = 0
            first_tile = -1
            last_tile = -1
            while col < (len(tiles)):
                if (tiles[col]%18) == wall_tile:
                    if first_tile == -1:
                        first_tile = col
                        if col == len(tiles) - 1:
                            first_tile = -1
                            col += 1
                        else:
                            col += 1
                    elif col == len(tiles) - 1:
                        last_tile = col
                        self.wall_vertices_for_Box2D.append(
                            {"type": wall_tile, "vertices": self.wall_vertices_h((first_tile, row), (last_tile, row))})
                        # self.wall_vertices_for_Box2D.append(self.wall_vertices_h((first_tile, row), (last_tile, row)))
                        for i in range(first_tile, last_tile + 1):
                            tiles[i] = 0
                        first_tile = -1
                        col += 1
                    else:
                        col += 1
                else:
                    if first_tile != -1:
                        last_tile = col - 1
                        if first_tile == last_tile:
                            first_tile = -1
                            col += 1
                        else:
                            self.wall_vertices_for_Box2D.append(
                                {"type": wall_tile,
                                 "vertices": self.wall_vertices_h((first_tile, row), (last_tile, row))})
                            for i in range(first_tile, last_tile + 1):
                                tiles[i] = 0
                            first_tile = -1
                            col += 1
                    else:
                        col += 1

    def wall_vertices_h(self, first_tile, last_tile):
        first_tilex = first_tile[0] + TILESIZE / (2 * PPM) + 1
        first_tiley = - first_tile[1] - TILESIZE / (2 * PPM) - 1
        last_tilex = last_tile[0] + TILESIZE / (2 * PPM) + 1
        last_tiley = - last_tile[1] - TILESIZE / (2 * PPM) - 1
        r = TILESIZE / (2 * PPM)
        vertices = [(first_tilex - r, first_tiley + r),
                    (last_tilex + r, last_tiley - r),
                    (last_tilex + r, last_tiley + r),
                    (first_tilex - r, first_tiley - r),

                    ]  # Box2D

        # self.wall_info.append([vertices[0],vertices[1]])
        # self.wall_info.append([vertices[2],vertices[1]])
        # self.wall_info.append([vertices[3],vertices[0]])
        # self.wall_info.append([vertices[2],vertices[3]])
        return vertices

    def wall_vertices_v(self, first_tile, last_tile):
        first_tilex = first_tile[0] + TILESIZE / (2 * PPM) + 1
        first_tiley = - first_tile[1] - TILESIZE / (2 * PPM) - 1
        last_tilex = last_tile[0] + TILESIZE / (2 * PPM) + 1
        last_tiley = - last_tile[1] - TILESIZE / (2 * PPM) - 1
        r = TILESIZE / (2 * PPM)
        vertices = [(first_tilex - r, first_tiley + r),
                    (first_tilex + r, first_tiley + r),
                    (last_tilex + r, last_tiley - r),
                    (last_tilex - r, last_tiley - r),

                    ]  # Box2D

        # self.wall_info.append([vertices[0],vertices[1]])
        # self.wall_info.append([vertices[2],vertices[1]])
        # self.wall_info.append([vertices[3],vertices[0]])
        # self.wall_info.append([vertices[2],vertices[3]])
        return vertices

    def _print_result(self):
        if self.is_end and self.x == 0:
            for user in self.ranked_user:

                self.result.append(str(user.car_no + 1) + "P:" + str(user.end_frame) + "frame")
            self.x += 1
            print(self.result)

    def load_map_object(self):
        for row, tiles in enumerate(self.map.data):
            for col, tile in enumerate(tiles):
                try:
                    if tile == 6 or tile == 10:
                        for world in self.worlds:
                            x, y = (col + (TILE_LEFTTOP[0] / TILESIZE), row + (TILE_LEFTTOP[1] / TILESIZE))
                            self.car = Car(world, (x + TILESIZE / (2 * PPM), - y - TILESIZE / (2 * PPM)),
                                           self.worlds.index(world), self.sensor_num, 2)
                            self.cars.add(self.car)
                            self.car_info.append(self.car.get_info())
                            # Car(self, world, (col + (TILE_LEFTTOP[0] / TILESIZE), row + (TILE_LEFTTOP[1] / TILESIZE)), i, 1)
                    elif tile == 7:
                        self.end_point = End_point(self,
                                                   (col + (TILE_LEFTTOP[0] / TILESIZE), row + (TILE_LEFTTOP[1] / TILESIZE)))
                        self.check_point_num += 1
                    elif tile == 8:
                        Check_point(self, (col + (TILE_LEFTTOP[0] / TILESIZE), row + (TILE_LEFTTOP[1] / TILESIZE)))
                        self.check_point_num+=1
                    elif tile == 9:
                        Outside_point(self, (col + (TILE_LEFTTOP[0] / TILESIZE), row + (TILE_LEFTTOP[1] / TILESIZE)))
                    elif tile == 13:
                        for world in self.worlds:
                            x, y = (col + (TILE_LEFTTOP[0] / TILESIZE), row + (TILE_LEFTTOP[1] / TILESIZE))
                            self.car = Car(world, (x + TILESIZE / (2 * PPM), - y - TILESIZE / (2 * PPM)),
                                           self.worlds.index(world), self.sensor_num, 0.5)
                            self.cars.add(self.car)
                            self.car_info.append(self.car.get_info())
                    elif tile == 12:
                        for world in self.worlds:
                            x, y = (col + (TILE_LEFTTOP[0] / TILESIZE), row + (TILE_LEFTTOP[1] / TILESIZE))
                            self.car = Car(world, (x + TILESIZE / (2 * PPM), - y - TILESIZE / (2 * PPM)),
                                           self.worlds.index(world), self.sensor_num, 1)
                            self.cars.add(self.car)
                            self.car_info.append(self.car.get_info())
                    elif tile == 11:
                        for world in self.worlds:
                            x, y = (col + (TILE_LEFTTOP[0] / TILESIZE), row + (TILE_LEFTTOP[1] / TILESIZE))
                            self.car = Car(world, (x + TILESIZE / (2 * PPM), - y - TILESIZE / (2 * PPM)),
                                           self.worlds.index(world), self.sensor_num, 1.5)
                            self.cars.add(self.car)
                            self.car_info.append(self.car.get_info())
                    elif tile == 14:
                        x, y = (col + (TILE_LEFTTOP[0] / TILESIZE), row + (TILE_LEFTTOP[1] / TILESIZE))
                        wall_vertices = [(x, -y),
                                         (x + (TILE_LEFTTOP[0] / TILESIZE), -y),
                                         (x + (TILE_LEFTTOP[0] / TILESIZE), -y - (TILE_LEFTTOP[0] / TILESIZE))]
                        for world in self.worlds:
                            wall = SlantWall(self, wall_vertices, world)
                            if self.worlds.index(world) == 0:
                                self.slant_walls.add(wall)
                    elif tile == 15:
                        x, y = (col + (TILE_LEFTTOP[0] / TILESIZE), row + (TILE_LEFTTOP[1] / TILESIZE))
                        wall_vertices = [(x, -y - (TILE_LEFTTOP[0] / TILESIZE)),
                                         (x + (TILE_LEFTTOP[0] / TILESIZE), -y),
                                         (x + (TILE_LEFTTOP[0] / TILESIZE), -y - (TILE_LEFTTOP[0] / TILESIZE))]
                        for world in self.worlds:
                            wall = SlantWall(self, wall_vertices, world)
                            if self.worlds.index(world) == 0:
                                self.slant_walls.add(wall)
                    elif tile == 16:
                        x, y = (col + (TILE_LEFTTOP[0] / TILESIZE), row + (TILE_LEFTTOP[1] / TILESIZE))
                        wall_vertices = [(x, -y - (TILE_LEFTTOP[0] / TILESIZE)),
                                         (x, -y),
                                         (x + (TILE_LEFTTOP[0] / TILESIZE), -y - (TILE_LEFTTOP[0] / TILESIZE))]
                        for world in self.worlds:
                            wall = SlantWall(self, wall_vertices, world)
                            if self.worlds.index(world) == 0:
                                self.slant_walls.add(wall)
                    elif tile == 17:
                        x, y = (col + (TILE_LEFTTOP[0] / TILESIZE), row + (TILE_LEFTTOP[1] / TILESIZE))
                        wall_vertices = [(x, -y - (TILE_LEFTTOP[0] / TILESIZE)),
                                         (x, -y),
                                         (x + (TILE_LEFTTOP[0] / TILESIZE), -y)]
                        for world in self.worlds:
                            wall = SlantWall(self, wall_vertices, world)
                            if self.worlds.index(world) == 0:
                                self.slant_walls.add(wall)
                except Exception:
                    print("Map Error")
        try:
            if self.end_point and len(self.cars):
                pass
            else:
                print("Map without car")
                self.running = False
        except:
            print("Map without end point")
            self.running = False




import pygame
from loguru import logger

from mlgame.view.audio_model import MusicInitSchema, SoundInitSchema


def mixer_enabled(func):
    def wrapper(self, *args, **kwargs):
        if self._is_mixer_enabled:
            return func(self, *args, **kwargs)

    return wrapper


class SoundController:
    def __init__(self, is_sound_on: bool, music_objs: list[MusicInitSchema], sound_objs: [SoundInitSchema]):
        self._is_mixer_enabled = is_sound_on

        self._music_path_dict = {}
        self._sound_dict = {}
        self._current_music_id = ''
        self._is_paused = False
        self._volume = 1
        try:
            if self._is_mixer_enabled:
                pygame.mixer.init()
                pygame.mixer.music.set_volume(1)
                # store music obj
                for music in music_objs:
                    self._music_path_dict[music.music_id] = music.file_path
                    # self.play_music(music.music_id)

                # store sound obj
                for sound in sound_objs:
                    self._sound_dict[sound.sound_id] = pygame.mixer.Sound(sound.file_path)

        except Exception as e:
            logger.error(e)
            self._is_mixer_enabled = False

    @mixer_enabled
    def toggle_sound(self):
        if self._volume == 0:
            self._volume = 1
        else:
            self._volume = 0
        pygame.mixer.music.set_volume(self._volume)
        
        

    @mixer_enabled
    def play_music(self, music_id: str):
        music_path = self._music_path_dict.get(music_id, None)
        #  music_id is playing , dont change

        if self._current_music_id != music_id and music_path:
            self._current_music_id = music_id
            pygame.mixer.music.load(self._music_path_dict[music_id])
            pygame.mixer.music.play(-1)

        pass

    @mixer_enabled
    def play_sound(self, sound_id):
        sound_obj = self._sound_dict.get(sound_id, None)
        if sound_obj and self._volume:
            self._sound_dict[sound_id].play()
        pass

    @mixer_enabled
    def pause(self):
        if not self._is_paused :
            pygame.mixer.music.pause()
            self._is_paused = True

    @mixer_enabled
    def unpause(self):
        if self._is_paused :
            pygame.mixer.music.unpause()
            self._is_paused = False



from os import path

'''width and height'''
WIDTH = 1000
HEIGHT = 700

'''environment data'''
FPS = 30
ceiling = 600
finish_line = 15000

'''color'''
BLACK = "#000000"
WHITE = "#ffffff"
RED = "#ff0000"
YELLOW = "#ffff00"
GREEN = "#00ff00"
GREY = "#8c8c8c"
BLUE = "#0000ff"
LIGHT_BLUE = "##21A1F1"

'''object size'''
car_size = (60, 30)
coin_size = (30,30)
lane_size = (20,3)

'''command'''
LEFT_cmd = "MOVE_LEFT"
RIGHT_cmd = "MOVE_RIGHT"
SPEED_cmd = "SPEED"
BRAKE_cmd = "BRAKE"

'''data path'''
ASSET_IMAGE_DIR = path.join(path.dirname(__file__), "../asset/image")
IMAGE_DIR = path.join(path.dirname(__file__), 'image')
SOUND_DIR = path.join(path.dirname(__file__), 'sound')
BACKGROUND_IMAGE = ["ground0.png"]
COIN_IMAGE = "coin.png"
RANKING_IMAGE = ["info_coin.png", "info_km.png"]

START_LINE_IMAGE = ["start.png", "finish.png"]
# FINISH_LINE_IMAGE =
USER_IMAGE = [["car-1.png","car1-bad.png"],["car-2.png","car2-bad.png"],
              ["car-3.png","car3-bad.png"], ["car-4.png","car4-bad.png"]]
COMPUTER_CAR_IMAGE = ["car-5.png","computer_die.png"]
USER_COLOR = [WHITE, YELLOW, BLUE, RED]

'''image url'''
COMPUTER_CAR_URL = "https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/RacingCar/main/asset/image/car-5.png"
USER_CAR_URL = ["https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/racing_car/main/asset/image/car-1.png",
                "https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/racing_car/main/asset/image/car-2.png",
                "https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/racing_car/main/asset/image/car-3.png",
                "https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/racing_car/main/asset/image/car-4.png"]
BACKGROUND_URL = "https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/racing_car/main/asset/image/ground0.png"
INFO_COIN_URL = "https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/racing_car/main/asset/image/info_coin.png"
INFO_KM_URL = "https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/racing_car/main/asset/image/info_km.png"
FINISH_URL = "https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/racing_car/main/asset/image/finish.png"
START_URL = "https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/racing_car/main/asset/image/start.png"
COIN_URL = "https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/racing_car/main/asset/image/coin.png"

computerCar_init_position = [
    (650, 120), (650, 170), (650, 220), (650, 270), (650, 320), (650, 370), (650, 420), (650, 470), (650, 520),
    (-700, 120), (-700, 170), (-700, 220), (-700, 270), (-700, 320), (-700, 370), (-700, 420), (-700, 470), (-700, 520)
]
# computerCar_init_position = [
#     (650, 110), (650, 160), (650, 210), (650, 260), (650, 310), (650, 360), (650, 410), (650, 460), (650, 510),
#     (-700, 110), (-700, 160), (-700, 210), (-700, 260), (-700, 310), (-700, 360), (-700, 410), (-700, 460), (-700, 510)
# ]

userCar_init_position = [170, 270, 370, 470]




import sys
from typing import Optional, Sequence
from mlgame.argument.cmd_argument import parse_cmd_and_get_arg_obj
from mlgame.mlgame_obj import run_mlgame
from mlgame.core.security import install_exec_killer


def main(argv: Optional[Sequence[str]] = None):
    argv = list(sys.argv[1:] if argv is None else argv)
    arg_obj = parse_cmd_and_get_arg_obj(argv)
    run_mlgame(arg_obj)

if __name__ == "__main__":
    import os
    os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = "hide"
    install_exec_killer()
    main()


import random
from os import path

import pygame

from mlgame.game.paia_game import GameStatus, GameResultState, PaiaGame
from mlgame.view.decorator import check_game_progress, check_game_result
from mlgame.view.view_model import create_text_view_data, Scene, create_scene_progress_data, create_asset_init_data, \
    create_image_view_data
from .env import BRICK_PATH, BRICK_URL, BALL_PATH, BOARD_PATH, BOARD_URL, BALL_URL, HARD_BRICK_PATH, HARD_BRICK_URL, \
    ASSET_LEVEL_DIR, BG_PATH, BG_URL, INFO_TXT_X, BG_LEFT_WIDTH
from .game_object import Ball, Platform, Brick, HardBrick, PlatformAction, SERVE_BALL_ACTIONS
from .utils import shift_left_with_bg_width, shift_left_with_bg_width_by_lru


class Arkanoid(PaiaGame):
    def __init__(self, user_num=1, level=1, level_file=None, *args, **kwargs):
        super().__init__(user_num=user_num)
        self.frame_count = 0

        def get_level_file_path(level: int, level_file: str):
            if level_file:
                level_file_path = level_file
            else:
                level_file_path = path.join(ASSET_LEVEL_DIR, f"{level}.dat")

            if not path.exists(level_file_path):
                print("level is not existed , turn to level 1")
                level_file_path = path.join(ASSET_LEVEL_DIR, "1.dat")
            return level_file_path

        self.level_file_path = get_level_file_path(level, level_file)

        self.game_result_state = GameResultState.UN_PASSED
        self.ball_served = False
        self.scene = Scene(width=1000, height=500, color="#54AFE3", bias_x=0, bias_y=0)
        self._hard_brick = []
        self._brick = []
        self._create_init_scene()

    def update(self, commands):
        ai_1p_cmd = commands[self.ai_clients()[0]["name"]]
        command = (PlatformAction(ai_1p_cmd)
                   if ai_1p_cmd in PlatformAction.__members__ else PlatformAction.NONE)

        self.frame_count += 1
        self._platform.move(command)

        if not self.ball_served:
            # Force to serve the ball after 150 frames
            if (self.frame_count >= 150 and
                    command not in SERVE_BALL_ACTIONS):
                command = random.choice(SERVE_BALL_ACTIONS)

            self._wait_for_serving_ball(command)
        else:
            self._ball_moving()

        if not self.is_running:
            return "RESET"

    def _wait_for_serving_ball(self, platform_action: PlatformAction):
        self._ball.stick_on_platform(self._platform.rect.centerx)

        if platform_action in SERVE_BALL_ACTIONS:
            self._ball.serve(platform_action)
            self.ball_served = True

    def _ball_moving(self):
        self._ball.move()

        hit_bricks, new_bricks = self._ball.check_hit_brick(self._group_brick)
        for brick in hit_bricks:
            if isinstance(brick, HardBrick):
                self._hard_brick.remove(brick)
            else:
                self._brick.remove(brick)
        self._brick.extend(new_bricks)

        self._ball.check_bouncing(self._platform)

    def get_data_from_game_to_player(self):
        to_players_data = {}
        data_to_1p = {
            "frame": self.frame_count,
            "status": self.get_game_status(),
            # TODO how to resolve different position
            "ball": shift_left_with_bg_width(self._ball.pos),
            "ball_served": self.ball_served,
            # "platform": self._platform.pos,
            "platform": shift_left_with_bg_width(self._platform.pos),
            "bricks": [],
            "hard_bricks": []
        }
        for brick in self._hard_brick:
            data_to_1p["hard_bricks"].append(shift_left_with_bg_width_by_lru(brick.pos))

        for brick in self._brick:
            data_to_1p["bricks"].append(shift_left_with_bg_width_by_lru(brick.pos))

        for ai_client in self.ai_clients():
            to_players_data[ai_client['name']] = data_to_1p

        return to_players_data


    def get_game_status(self):
        if len(self._group_brick) == 0:
            self._game_status = GameStatus.GAME_PASS
        elif self._ball.rect.top >= self._platform.rect.bottom:
            self._game_status = GameStatus.GAME_OVER
        elif self._ball.hit_brick_false > 50:
            self._game_status = GameStatus.GAME_PASS
        else:
            self._game_status = GameStatus.GAME_ALIVE
        return self._game_status

    def reset(self):
        self.game_result_state = GameResultState.UN_PASSED
        self.ball_served = False
        self.frame_count = 0
        self._brick = []
        self._hard_brick = []
        self._create_init_scene()
        pass

    @property
    def is_running(self):
        return self.get_game_status() == GameStatus.GAME_ALIVE

    def get_scene_init_data(self):

        scene_init_data = {
            "scene": self.scene.__dict__,
            "assets": [
                create_asset_init_data("brick", 25, 10, BRICK_PATH, BRICK_URL),
                create_asset_init_data("hard_brick", 25, 10, HARD_BRICK_PATH, HARD_BRICK_URL),
                create_asset_init_data("ball", 5, 5, BALL_PATH, BALL_URL),
                create_asset_init_data("board", 40, 5, BOARD_PATH, BOARD_URL),
                create_asset_init_data("bg", 1000, 500, BG_PATH, BG_URL),
            ],
            "background": [
                create_image_view_data("bg", 0, 0, 1000, 500),
            ]
        }
        return scene_init_data

    @check_game_progress
    def get_scene_progress_data(self):
        bricks_data = []
        lines = []
        for brick in self._group_brick:
            bricks_data.append(brick.get_object_data)
            # lines.append(brick.get_line_data1)
            # lines.append(brick.get_line_data2)

        game_obj_list = []
        for move in self._group_move:
            game_obj_list.append(move.get_object_data)

        game_obj_list.extend(bricks_data)
        game_obj_list.extend(lines)

        catch_ball_text = create_text_view_data(
            "catching ball: " + str(self._ball.hit_platform_times), INFO_TXT_X,
            self.scene.height - 21, "#FFFFFF", "16px Arial")

        remain_brick_text = create_text_view_data(
            "remain brick: " + str(len(self._brick)), INFO_TXT_X,
            self.scene.height - 41, "#FFFFFF", "16px Arial")
        remain_hard_brick_text = create_text_view_data(
            "remain hard brick: " + str(len(self._hard_brick)), INFO_TXT_X,
            self.scene.height - 61, "#FFFFFF", "16px Arial")
        foreground = [catch_ball_text, remain_brick_text, remain_hard_brick_text]
        # foreground.extend(lines)

        scene_progress = create_scene_progress_data(
            frame=self.frame_count, object_list=game_obj_list,
            foreground=foreground)
        return scene_progress

    @check_game_result
    def get_game_result(self):
        if self._game_status == GameStatus.GAME_PASS:
            self.game_result_state = GameResultState.PASSED
        return {
            "frame_used": self.frame_count,
            "status": self.game_result_state,
            "attachment": [
                {
                    "player_num": self.ai_clients()[0]['name'],
                    "rank": 1,
                    "brick_remain": len(self._brick) + 2 * len(self._hard_brick),
                    "count_of_catching_ball": self._ball.hit_platform_times

                }
            ]

        }

    def get_keyboard_command(self):
        cmd_1p = "NONE"
        key_pressed_list = pygame.key.get_pressed()
        if key_pressed_list[pygame.K_a]:
            cmd_1p = "SERVE_TO_LEFT"
        elif key_pressed_list[pygame.K_d]:
            cmd_1p = "SERVE_TO_RIGHT"
        elif key_pressed_list[pygame.K_LEFT]:
            cmd_1p = "MOVE_LEFT"
        elif key_pressed_list[pygame.K_RIGHT]:
            cmd_1p = "MOVE_RIGHT"
        else:
            cmd_1p = "NONE"

        ai_1p = self.ai_clients()[0]["name"]

        return {ai_1p: cmd_1p}

    def _create_init_scene(self):
        '''
        初始遊戲畫面：
        1. 球
        2. 板子
        3. 磚塊
        '''
        self._create_moves()

        self._create_bricks()

    def _create_moves(self):
        self._group_move = pygame.sprite.RenderPlain()
        # TODO revise init pos
        self._ball = Ball((BG_LEFT_WIDTH+93, 391), pygame.Rect(BG_LEFT_WIDTH, 0, 200, 500), self._group_move)
        self._platform = Platform((BG_LEFT_WIDTH+75, 400), pygame.Rect(BG_LEFT_WIDTH, 0, 200, 500), self._group_move)

    def _create_bricks(self):
        def get_coordinate_and_type(string):
            string = string.rstrip("\n").split(' ')
            return int(string[0]), int(string[1]), int(string[2])

        self._group_brick = pygame.sprite.RenderPlain()
        self._brick_container = []

        with open(self.level_file_path, 'r') as input_file:
            offset_x, offset_y, _ = get_coordinate_and_type(input_file.readline())
            for input_pos in input_file:
                pos_x, pos_y, type = get_coordinate_and_type(input_pos.rstrip("\n"))
                BrickType = {
                    0: Brick,
                    1: HardBrick,
                }.get(type, Brick)
                # TODO revise position
                brick = BrickType((BG_LEFT_WIDTH + pos_x + offset_x, pos_y + offset_y),
                                  self._group_brick)
                self._brick_container.append(brick)

                if BrickType == Brick:
                    self._brick.append(brick)
                else:
                    self._hard_brick.append(brick)

    @staticmethod
    def ai_clients():
        """
        let MLGame know how to parse your ai,
        you can also use this names to get different cmd and send different data to each ai client
        """
        return [
            {"name": "1P"}
        ]



import json
from pprint import pprint

import pygame


class MLPlay:
    def __init__(self, ai_name,*args,**kwargs):
        self.player_no = ai_name
        self.r_sensor_value = 0
        self.l_sensor_value = 0
        self.f_sensor_value = 0
        self.control_list = {"left_PWM": 0, "right_PWM": 0}
        # print("Initial ml script")

    def update(self, scene_info: dict, keyboard: list = [], *args, **kwargs):
        """
        Generate the command according to the received scene information
        """
        print(json.dumps(scene_info,indent=2))
        if scene_info["status"] != "GAME_ALIVE":
            return "RESET"
        if pygame.K_w in keyboard or pygame.K_UP in keyboard:
            self.control_list["left_PWM"] = 200
            self.control_list["right_PWM"] = 200
        elif pygame.K_a in keyboard or pygame.K_LEFT in keyboard:
            self.control_list["left_PWM"] = -150
            self.control_list["right_PWM"] = 150
        elif pygame.K_d in keyboard or pygame.K_RIGHT in keyboard:
            self.control_list["left_PWM"] = 150
            self.control_list["right_PWM"] = -150
        elif pygame.K_s in keyboard or pygame.K_DOWN in keyboard:
            self.control_list["left_PWM"] = -150
            self.control_list["right_PWM"] = -150
        else:

            self.control_list["left_PWM"] = 100
            self.control_list["right_PWM"] = 100

        return self.control_list

    def reset(self):
        """
        Reset the status
        """
        # print("reset ml script")
        pass



import copy
import json

from mlgame.argument.model import UserNumConfig, GroupAI, AINameEnum
from mlgame.utils.logger import logger


def get_data_from_json_file(file_path) -> dict:
    """
    open json file and return dict data
    """
    with open(file=file_path, mode="rb") as f:
        config_data = json.load(f)
    return config_data


def revise_ai_clients(ai_clients: list[GroupAI], user_num_config: UserNumConfig)->list[GroupAI]:
    ai_clients_result = ai_clients.copy()
    if len(ai_clients) < user_num_config.min:
        logger.warning("提供的ＡＩ數量小於遊戲最小所需ＡＩ數量，系統將會使用最後一個ＡＩ自動補足")
        index =2
        while len(ai_clients_result) < user_num_config.min:

            ai_client = copy.deepcopy(ai_clients[-1])
            ai_client.ai_name=AINameEnum(f"{index}P")
            ai_clients_result.append(ai_client)
            index += 1
    elif len(ai_clients) > user_num_config.max:
        logger.warning("提供的ＡＩ數量高過遊戲容許範圍，系統將會忽略倒數幾個ＡＩ")
        ai_clients_result = ai_clients[:user_num_config.max]
    return ai_clients_result



import pygame
from .env import *

class Coin(pygame.sprite.Sprite):
    def __init__(self, x, y):
        pygame.sprite.Sprite.__init__(self)
        self.rect = pygame.Rect(0, 0, coin_size[0], coin_size[1])
        self.rect.center = x ,y
        self.vel = 5

    def update(self, *args):
        self.rect.centerx -= self.vel
        if self.rect.centerx < -500:
            self.kill()

    def move(self):
        self.rect.centery += self.vel

    def get_position(self):
        return (self.rect.left, self.rect.top)



"""
This is a base class for different mode in src.
"""
import pygame
from .env import *
from .car import Camera, UserCar, ComputerCar
from .highway import Lane, Line
import random
from mlgame.game.paia_game import GameResultState, GameStatus

class GameMode(object):
    def __init__(self, user_num: int, car_num, length:int, sound_controller):
        self.bg_rect = pygame.Rect(0, 0, 2000, HEIGHT)
        self.running = True
        self.frame = 0
        '''sound'''
        self.sound_controller = sound_controller

        '''groups'''
        self.users = pygame.sprite.Group()
        self.cars = pygame.sprite.Group()
        self.computerCars = pygame.sprite.Group()
        self.lanes = pygame.sprite.Group()
        self.camera = Camera(length)

        '''data set'''
        random.shuffle(userCar_init_position)
        self.cars_num = car_num
        self.background_x = 0
        self.rel_x = 0
        self.bg_x = 0
        self.maxVel = 0
        self.user_distance = []
        self.eliminated_user = []
        self.winner = []
        if user_num == 1:
            self.is_single = True
        else:
            self.is_single = False

        self.length = length
        '''initial object'''
        for user in range(user_num):
            self._init_user(user)
        self.line = Line(self.length)
        self.lanes.add(self.line)
        self._init_lanes()

        '''state include GameResultState.FINISH、GameResultState.FAIL"'''
        self.state = GameResultState.FAIL
        self.last_create_car = 0

    def handle_event(self):
        """ Handle the event from window , mouse or button.
        :return: None
        """
        pass

    def detect_collision(self):
        """ Detect the collision event between sprites.
        :return: None
        """
        pass

    def update(self, command):
        """ This function should update every sprite in games.
        :return: None
        """
        pass

    def count_bg(self):
        """  Draw a background on screen.
        :return:None
        """
        self.rel_x = self.background_x % self.bg_rect.width
        self.bg_x = self.rel_x - self.bg_rect.width
        self.background_x -= self.camera.velocity

    def isRunning(self) -> bool:
        return self.running

    def _init_user(self, user_no: int):
        self.car = UserCar(userCar_init_position[user_no], 0, user_no)
        self.users.add(self.car)
        self.cars.add(self.car)

    def _init_lanes(self):
        for i in range(8):
            for j in range(23):
                self.lane = Lane(i * 50+150, j * 50-150)
                self.lanes.add(self.lane)

    def _creat_computercar(self):
        if len(self.cars) < self.cars_num and self.frame - self.last_create_car >10:
            for i in range(2):
                x, y = random.choice(computerCar_init_position)
                computerCar = ComputerCar(y, self.camera.position + x, x + 500)
                self.computerCars.add(computerCar)
                self.cars.add(computerCar)
            self.last_create_car = self.frame
            # x = random.choice([650, -700])
            # y = random.choice(self.car_lanes)

    def user_out_screen(self,car):
        if car.state:
            if car.rect.right < -100 or car.rect.bottom > 540 or car.rect.top < 100:
                self.sound_controller.play_lose_sound()
                car.state = False

    def _detect_car_status(self, car):
        if car.state:
            pass
        else:
            car.velocity = 0
            if car in self.users:
                if car not in self.eliminated_user:
                    car.status = GameStatus.GAME_OVER
                    self.eliminated_user.append(car)
                    self.user_distance.append(car.distance)
            else:
                car.kill()

    def _revise_speed(self):
        self.user_vel = []
        for car in self.users:
            if not car.state:
                self.maxVel = 0
                # return 0
            self.user_vel.append(car.velocity)
        self.maxVel = max(self.user_vel)



import os
import sys
from argparse import ArgumentParser, REMAINDER

import pydantic

from mlgame.argument.model import MLGameArgument
from mlgame.utils.logger import logger
from mlgame.version import version


def create_cli_args_parser():
    """
    Generate an ArgumentParser for parse the arguments in the command line
    """
    usage_str = "python -m mlgame [options] <game_folder> [game_params]"
    description_str = ("A platform for applying machine learning algorithm "
                       "to play pixel games. "
                       "In default, the game runs in the machine learning mode. ")

    parser = ArgumentParser(usage=usage_str, description=description_str,
                            add_help=False)

    parser.add_argument("game_folder",
                        type=os.path.abspath,
                        nargs="?",
                        help="the name of the game to be started")
    parser.add_argument("game_params", nargs=REMAINDER, default=None,
                        help="[optional] the additional settings for the game. "
                             "Note that all arguments after <game> will be collected to 'game_params'.")

    group = parser.add_argument_group(title="functional options")
    group.add_argument("-v", "--version", action="version", version=version)
    group.add_argument("-h", "--help", action="help",
                       help="show this help message and exit. "
                            "If this flag is specified after the <game>, "
                            "show the help message of the game instead.")

    group.add_argument("-f", "--fps", type=int, default=30,
                       help="the updating frequency of the game process [default: %(default)s]")

    group.add_argument("-1", "--one-shot", action="store_true",
                       dest="one_shot_mode",
                       help="quit the game when the game is passed or is over. "
                            "Otherwise, the game will restart automatically. [default: %(default)s]")
    group.add_argument("--nd", "--no-display", action="store_true",
                       dest="no_display", default=False,
                       help="didn't display the game on screen. [default: %(default)s]")
    group.add_argument("--ws_url",
                       type=str,
                       dest="ws_url",
                       help="ws_url route")

    group.add_argument("-i", "--input-ai",
                       # type=validate_file,
                       type=str,
                       action="append",
                       dest="labeled_ai_orig",
                       default=None, metavar="SCRIPT",
                       help="specify user script(s) for the machine learning mode. "
                            "For multiple user scripts, use this flag multiple times. "
                            "The script path could be relative path or absolute path "
                       )
    group.add_argument("-o", "--output-folder",
                       # type=validate_file,
                       type=os.path.abspath,
                       dest="output_folder",
                       default=None, metavar="SCRIPT",
                       help="Save each frame of game progress into destination folder."
                            "This folder should be existed and it will create a timestamp folder."
                       )
    group.add_argument("-r", "--progress-folder","--record-folder",
                       # type=validate_file,
                       type=os.path.abspath,
                       dest="record_folder",
                       default=None, metavar="SCRIPT",
                       help="Save game process as a file into destination folder."
                            "This folder should be existed and it will create a timestamp folder."
                       )

    group.add_argument("-p", "--progress-frame-frequency", type=int, default=300,
                       help="the frequency of the game progress save [default: %(default)s]")
    group.add_argument("--ns", "--no-sound", action="store_false",
                       dest="is_sound_on", default=True,
                       help="didn't play music and sound. [default: %(default)s]")
    group.add_argument("--debug", action="store_true",
                       dest="is_debug", default=False,
                       help="show all debug info in console. [default: %(default)s]")

    # group.add_argument("--group_ai",default=None,
    #                    action="append",
    #                    dest="group_ai_orig",
    #                    help="set ai with team group, team should be one of ['A','B','C','D'] ,"
    #                         "player number should be ['1P','2P',...,'8P'] \n"
    #                         "!Important: this argument will overwrite input from --input-ai or -i \n"
    #                         "format: --group_ai A,1P,/path/to/ml_play.py,ai_label "
    #                    )
    group.add_argument("--az_upload_url",
                       type=str,
                       dest="az_upload_url",
                       help="Azure Blob Storage upload URL")

    return parser


def parse_cmd_and_get_arg_obj(arg_str: list) -> MLGameArgument:
    arg_parser = create_cli_args_parser()
    try:
        parsed_args = arg_parser.parse_args(arg_str)
    except pydantic.ValidationError as e:
        logger.exception(f"Error in parsing command : {e.__str__()}")
        arg_parser.print_help()
        sys.exit()

    arg_obj = MLGameArgument(**parsed_args.__dict__)
    # how to resolve if file or directory at path  does not exist
    return arg_obj



from .car import *
from .highway import *
from .gameMode import GameMode
from .env import *
import pygame
import random
from mlgame.game.paia_game import GameResultState, GameStatus


class ReliveMode(GameMode):
    def __init__(self, user_num: int, car_num, length, sound_controller):
        super(ReliveMode, self).__init__(user_num, car_num, length, sound_controller)
        self.car_arrived = 0
        self.user_frames = [] # 使用者抵達終點所使用的時間
        self.limit_frame = length/1000 * 300

    def update(self, command):
        '''update the model of src,call this fuction per frame'''
        self.count_bg()
        self.frame += 1
        self.handle_event()
        self._revise_speed() # get the velocity of user, and revise max_vel

        if self.frame > FPS:
            if self.frame > FPS * 4:
                self._creat_computercar()

            self.camera.update(self.maxVel)

            '''update sprite'''
            self.lanes.update(self.camera.position)

            for car in self.users:
                self.user_out_screen(car)
                car.update(command[str(car.car_no + 1) + "P"])

            for car in self.cars:
                '''偵測車子的狀態'''
                self._detect_car_status(car)
                '''更新車子位置'''
                car.rect.left = car.distance - self.camera.position + 520
            self.computerCars.update(self.cars)

        if self._is_game_end(self.car_arrived):
            self.rank()
            self._print_result()
            self.running = False

    def _detect_car_status(self, car):
        super(ReliveMode, self)._detect_car_status(car)
        if car.state:
            pass
        else:
            if car in self.users:
                car.velocity = 0
            else:
                car.kill()

    def detect_collision(self):
        super(ReliveMode, self).detect_collision()
        for car in self.cars:
            self.cars.remove(car) # 如果sprite本身也在要偵測的group裡面就會被偵測到
            hits = pygame.sprite.spritecollide(car, self.cars, False)
            for hit in hits:
                if (hit.state and 0 < hit.rect.centerx < WIDTH):
                    self.sound_controller.play_hit_sound()
                if car in self.users:
                    if self.frame - car.cash_frame > 3*FPS:
                        car.velocity = 0
                        car.cash_frame = self.frame
                    # car.status = GameStatus.GAME_OVER
                else:
                    car.state = False
                if hit in self.users:
                    if self.frame - hit.cash_frame > 3*FPS:
                        hit.velocity = 0
                        hit.cash_frame = self.frame
                    # hit.status = GameStatus.GAME_OVER
                else:
                    hit.state = False
            self.cars.add(car)

    def _print_result(self):
        '''
        依照排名順序印出玩家遊戲結果，以字典形式顯示，包含以下項目：
        'player':標示玩家代號，顯示為1P、2P等
        'distance':記錄家此局所行走的距離
        'rank':顯示完家此局排名
        :return:None
        '''
        tem = []
        for user in self.winner:
            tem.append({"player_num": str(user.car_no + 1) + "P",
                        "distance": str(round(user.distance)) + "m",
                        "used_frames": str(user.used_frame)+" frames",
                        "single_rank": self.winner.index(user) + 1
                        })
            print({"player_num": str(user.car_no + 1) + "P",
                   "distance": str(round(user.distance)) + "m",
                    "used_frames": str(user.used_frame)+" frames",
                   "single_rank":self.winner.index(user)+1
                   })
        self.winner = tem

    def _is_game_end(self, car_arrived:int):
        '''
        判斷遊戲是否結束，遊戲結束的條件如下：
        1. 所有玩家皆抵達終點
        2. 首位玩家通過終點後再1000 frames
        :return:Bool
        '''
        end = True
        if self.frame >= self.limit_frame:
            return True
        if car_arrived == 0: # 尚未有車子通過終點
            for car in self.users:
                if self._is_car_arrive_end(car):
                    self.car_arrived = self.frame
        else:
            if self.frame - car_arrived > 1000:# 有玩家通過終點且已經過1000 frames
                return True
            else:
                for car in self.users:
                    if not self._is_car_arrive_end(car):
                        end = False
                if end:
                    return True
                else:
                    return False


    def _is_car_arrive_end(self, car):
        '''
        :param car: User
        :return: Bool
        '''
        if car.distance >= self.length:
            car.distance = self.length
            car.state = False
            return True
        return False

    def rank(self):
        '''
        如果玩家有通過終點則標示為GAME_PASS，反之為GAME_OVER
        排名先依據抵達終點所費之時間，若未抵達終點則以行進距離較遠者排名靠前
        '''
        self.user_distance = []
        for user in self.users:
            if user.distance >= self.length:
                user.status = GameStatus.GAME_PASS
                self.user_frames.append(user.used_frame)
            else:
                user.status = GameStatus.GAME_OVER
                self.user_distance.append(user.distance)
        while len(self.user_frames) > 0:
            for car in self.users:
                if self.user_frames:
                    if car.used_frame == min(self.user_frames):
                        self.winner.append(car)
                        self.user_frames.remove(car.used_frame)
        while len(self.user_distance) > 0:
            for car in self.users:
                if self.user_distance:
                    if car.distance == max(self.user_distance):
                        self.winner.append(car)
                        self.user_distance.remove(car.distance)
        # self.winner.reverse()

    def user_out_screen(self,car):
        if car.state:
            if car.rect.bottom > 540:
                if self.frame - car.cash_frame > 3 * FPS:
                    car.cash_frame = self.frame
                    car.velocity = 0
                car.rect.bottom = 540
            elif car.rect.top < 100:
                if self.frame - car.cash_frame > 3 * FPS:
                    car.velocity = 0
                    car.cash_frame = self.frame
                car.rect.top = 100





from typing import Optional
import pygame.sprite

from src.Player import Player
from src.Bullet import Bullet
from src.Wall import Wall


def collide_with_walls(group1: pygame.sprite.Group, group2: pygame.sprite.Group):
    hits = pygame.sprite.groupcollide(group1, group2, False, False, pygame.sprite.collide_rect_ratio(0.8))
    for sprite, walls in hits.items():
        sprite.collide_with_walls()


def collide_with_bullets(group1: pygame.sprite.Group, group2: pygame.sprite.Group, green_team_num: Optional[int] = None):
    hits = pygame.sprite.groupcollide(group1, group2, False, False, pygame.sprite.collide_rect_ratio(0.8))
    player_score_data = {}
    for sprite, bullets in hits.items():
        for bullet in bullets:
            if bullet.no != sprite.no and sprite.lives > 0:
                bullet.kill()

                if isinstance(sprite, Player):
                    assert green_team_num is not None
                    if (bullet.no - 1) // green_team_num == (sprite.no - 1) // green_team_num:
                        # -20 for friendly damage
                        score = -20
                    else:
                        score = 20
                elif isinstance(sprite, Wall):
                    score = 1
                    if sprite.lives == 1:
                        score += 5
                else:
                    continue

                sprite.collide_with_bullets()

                if player_score_data.get(bullet.no) is None:
                    player_score_data[bullet.no] = 0
                player_score_data[bullet.no] += score
    return player_score_data


def collide_with_supply_stations(sprites: pygame.sprite.Group, supply_stations: pygame.sprite.Group):
    hits = pygame.sprite.groupcollide(sprites, supply_stations, False, False, pygame.sprite.collide_rect_ratio(0.8))
    for sprite, supply_station in hits.items():
        if isinstance(sprite, Player):
            if supply_station[0].id == 5:
                sprite.get_oil(supply_station[0].power)
            else:
                sprite.get_power(supply_station[0].power)
        elif isinstance(sprite, Bullet):
            sprite.kill()

        supply_station[0].collect()

    return [supply_station[0] for supply_station in hits.values()]



from .coin import Coin
from .car import *
from .highway import *
from .gameMode import GameMode
from .env import *
import pygame
import random
from mlgame.game.paia_game import GameResultState, GameStatus


class CoinMode(GameMode):
    def __init__(self, user_num: int, car_num, length, sound_controller):
        super(CoinMode, self).__init__(user_num, car_num, length, sound_controller)
        self.coins = pygame.sprite.Group()
        self.is_arrive = False

        '''image'''
        self.create_coin_frame = 0
        self.coin_lanes = [125, 175, 225, 275, 325, 375, 425, 475, 525]

    def update(self, command):
        '''update the model of src,call this fuction per frame'''
        self.count_bg()
        self.frame += 1
        self.handle_event()
        self._revise_speed()

        if self.frame > FPS:
            if self.frame > FPS * 3:
                self._creat_computercar()
            if self.is_create_coin():
                self.create_coins()
            self.user_distance = []
            self.coin_num = []

            self.camera.update(self.maxVel)

            '''update sprite'''
            self.lanes.update(self.camera.position)
            self.coins.update()
            self.computerCars.update(self.cars)

            for car in self.users:
                self.user_out_screen(car)
                self.coin_num.append(car.coin_num)
                car.update(command[str(car.car_no + 1) + "P"])
                if self._is_car_arrive_end(car):
                    self.is_arrive = True
                    self.eliminated_user.append(car)
                    self.user_distance.append(car.distance)
                    break # 任一玩家通過終點則結束遊戲

            for car in self.cars:
                '''偵測車子的狀態'''
                self._detect_car_status(car)

                '''更新車子位置'''
                car.rect.left = car.distance - self.camera.position + 520

            if self._is_game_end(self.is_arrive):
                self.rank()
                self._print_result()
                self.running = False

    def detect_collision(self):
        super(CoinMode, self).detect_collision()
        for car in self.cars:
            self.cars.remove(car)  # 如果sprite本身也在要偵測的group裡面就會被偵測到
            hits = pygame.sprite.spritecollide(car, self.cars, False)
            for hit in hits:
                if (hit.state and 0 < hit.rect.centerx < WIDTH):
                    self.sound_controller.play_hit_sound()
                hit.state = False
                car.state = False
            self.cars.add(car)
        for car in self.users:
            hits = pygame.sprite.spritecollide(car, self.coins, True)
            for hit in hits:
                self.sound_controller.play_coin_sound()
                car.coin_num += 1

    def _print_result(self):
        '''
        依照排名順序印出玩家遊戲結果，以字典形式顯示，包含以下項目：
        'player':標示玩家代號，顯示為1P、2P等
        'coin':記錄玩家此局獲得的金幣數量
        'distance':記錄家此局所行走的距離
        'rank':顯示完家此局排名
        :return:None
        '''
        tem = []
        for user in self.winner:
            tem.append({
                "player_num": str(user.car_no + 1) + "P",
                "coin": str(user.coin_num),
                "used_frames": str(user.used_frame)+" frames",
                "distance": str(round(user.distance)) + "m",
                "single_rank": self.winner.index(user) + 1
            })
            print({
                "player_num": str(user.car_no + 1) + "P",
                "coin": str(user.coin_num),
                "used_frames": str(user.used_frame)+" frames",
                "distance": str(round(user.distance)) + "m",
                "single_rank": self.winner.index(user) + 1
            })
        self.winner = tem

    def _is_game_end(self, is_arrive:bool):
        '''
        判斷遊戲是否結束，遊戲結束的條件如下：
        單人模式：抵達終點，或是玩家出局(FAIL)
        多人模式：一名或以上玩家抵達終點，或市場上僅餘一名玩家
        :return:Bool
        '''
        if is_arrive:
            if self.is_single:
                self.state = GameResultState.PASSED
            else:
                self.state = GameResultState.FINISH
            return True
        if self.is_single:
            if len(self.eliminated_user) == 1:
                self.state = GameResultState.UN_PASSED
                return True

            return False
        else:
            if len(self.users) <= len(self.eliminated_user):
                self.state = GameResultState.FINISH
                return True
            else:
                return False

    def _is_car_arrive_end(self, car):
        '''
        :param car: User
        :return: Bool
        '''
        if car.distance > self.length:
            car.status = GameStatus.GAME_PASS
            return True
        return False

    def rank(self):
        user_value = []
        for car in self.users:
            if car not in self.eliminated_user:
                car.status = GameStatus.GAME_PASS
                self.eliminated_user.append(car)
            user_value.append(car.coin_num * 100000 + car.distance)
        while len(self.eliminated_user) != 0:
            for car in self.eliminated_user:
                car_value = car.coin_num * 100000 + car.distance
                if car_value == max(user_value):
                    self.winner.append(car)
                    user_value.remove(car_value)
                    self.eliminated_user.remove(car)

    def create_coins(self):
        if self.frame - self.create_coin_frame > FPS * 1.5:
            coin = Coin(WIDTH, random.choice(self.coin_lanes))
            self.coin_lanes.remove(coin.rect.centery)
            self.coins.add(coin)
            self.create_coin_frame = self.frame
        if len(self.coin_lanes) == 0:
            self.coin_lanes = [125, 175, 225, 275, 325, 375, 425, 475, 525]
        else:
            pass

    def is_create_coin(self):
        if self.maxVel >= 8:
            return True
        else:
            return False



import sys

import pygame

sys.path.append(r"../..")

from mlgame.view.view import PygameView
from mlgame.game.generic import quit_or_esc
from src.game import Arkanoid

FPS = 30
if __name__ == '__main__':
    pygame.init()
    game = Arkanoid( level=15)
    scene_init_info_dict = game.get_scene_init_data()
    game_view = PygameView(scene_init_info_dict)
    frame_count = 0
    while game.is_running and not quit_or_esc():
        pygame.time.Clock().tick_busy_loop(FPS)
        commands = game.get_keyboard_command()
        game.update(commands)
        game_progress_data = game.get_scene_progress_data()
        game_view.draw(game_progress_data)
        frame_count += 1

    pygame.quit()



from mlgame.view.decorator import check_scene_init_data, check_game_progress
from mlgame.view.view_model import get_scene_init_sample_data, get_dummy_progress_data


@check_scene_init_data
def example_of_use_init_decorator():
    """
    add @check_init_data will check the returned result of this function
    if you want to validate your init data, you could add this decorator easily to check
    :return:
    """
    # collect your init data here
    result = get_scene_init_sample_data()
    return result


@check_game_progress
def example_of_use_progress_decorator():
    """
    add @check_init_data will check the returned result of this function
    if you want to validate your init data, you could add this decorator easily to check
    :return:
    """
    # collect your init data here
    result = get_dummy_progress_data()
    return result


def test_example():
    example_of_use_init_decorator()
    example_of_use_progress_decorator()



import pytmx


def create_construction(_id: int or str, _no: int, _init_pos: tuple, _init_size: tuple):
    return {
        "_id": _id
        , "_no": _no
        , "_init_pos": _init_pos
        , "_init_size": _init_size
    }


# Map 讀取地圖資料
class TiledMap:
    def __init__(self, filepath: str):
        tm = pytmx.TiledMap(filepath)
        self.tile_width = tm.tilewidth
        self.tile_height = tm.tileheight
        self.width = tm.width
        self.height = tm.height
        self.map_width = self.tile_width * self.width
        self.map_height = self.tile_height * self.height
        self.tmx_data = tm
        self._is_record = False
        self.all_pos_list = []
        self.empty_pos_list = []
        self.empty_quadrant_pos_dict = {1: [], 2: [], 3: [], 4: []}
        self.all_obj_data_dict = {}
        # TODO refactor
        self.all_obj = {}

    def add_init_obj_data(self, img_id: int, cls, **kwargs):
        obj_data = {img_id: {"cls": cls,
                             "kwargs": kwargs
                             }
                    }
        self.all_obj_data_dict.update(obj_data)
        self.all_obj[img_id] = []

    def create_init_obj_dict(self) -> dict:
        obj_no = 0
        for layer in self.tmx_data.visible_layers:
            for x, y, gid, in layer:
                if isinstance(layer, pytmx.TiledTileLayer):
                    pos = (x * self.tile_width, y * self.tile_height)
                    if not self._is_record:
                        self.all_pos_list.append(pos)
                    if not self._is_record and not gid:  # 0代表空格，無圖塊
                        self.empty_pos_list.append(pos)
                        if pos[0] >= self.map_width // 2 and pos[1] < self.map_height // 2:
                            self.empty_quadrant_pos_dict[1].append(pos)
                        elif pos[0] < self.map_width // 2 and pos[1] < self.map_height // 2:
                            self.empty_quadrant_pos_dict[2].append(pos)
                        elif pos[0] < self.map_width // 2 and pos[1] >= self.map_height // 2:
                            self.empty_quadrant_pos_dict[3].append(pos)
                        else:
                            self.empty_quadrant_pos_dict[4].append(pos)
                    elif gid:
                        img_id = layer.parent.tiledgidmap[gid]
                        kwargs = self.all_obj_data_dict[img_id]["kwargs"]
                        obj_no += 1
                        img_info = {"_id": img_id, "_no": obj_no
                            , "_init_pos": pos
                            , "_init_size": (self.tile_width, self.tile_height)
                                    }
                        self.all_obj[img_id].append(self.all_obj_data_dict[img_id]["cls"](img_info, **kwargs))
        self._is_record = True
        return self.all_obj



from os import path

from .version import version

PPM = 16.0  # pixels per meter
TARGET_FPS = 60
TIME_STEP = 1.0 / TARGET_FPS

'''width and height'''
WIDTH = 800
HEIGHT = 640
TILE_WIDTH = 540  # 大小
TILE_HEIGHT = 540

'''tile-base'''
TILESIZE = 16
TILE_LEFTTOP = 16, 16  # pixel
GRIDWIDTH = (TILE_WIDTH + TILE_LEFTTOP[0]) / TILESIZE
GRIDHEIGHT = (TILE_HEIGHT + TILE_LEFTTOP[1]) / TILESIZE

'''sensor set trans'''
sensor_trans = ((1, 0),
                (0, 1),
                (-1, 0),
                (0, -1))

'''environment data'''
FPS = 30

'''color'''
GRAY = "#cccccc"
WHITE = "#ffffff"
BLACK = "#000000"
RED = "#C3305b"
YELLOW = "#f5d750"
BLUE = "#3a84c1"
LIGHT_BLUE = "#061c42"
PAIA_B = "#0c3997"
GREEN = "#50aa82"
SENSOR_Y = "#ffff83"
SENSOR_R = "#ed2323"
SENSOR_B = "#1d92fe"
CAR_COLOR = [RED, BLUE, GREEN, YELLOW]

'''object size'''
car_size = (60, 30)

'''data path'''
ASSET_IMAGE_DIR = path.join(path.dirname(__file__), "../asset/image")
IMAGE_DIR = path.join(path.dirname(__file__), 'image')
SOUND_DIR = path.join(path.dirname(__file__), '../asset/sound')
MUSIC_DIR = path.join(path.dirname(__file__), '../asset/music')
MAP_DIR = path.join(path.dirname(__file__), "map")
IMAGE_URL = f"https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/dont_touch/{version}/asset/image/"
SOUND_URL = f"https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/dont_touch/{version}/asset/sound/"
MUSIC_URL = f"https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/dont_touch/{version}/asset/music/"
'''image'''
BG_IMG = "bg.png"
BG_URL = f"{IMAGE_URL}bg.png"

L_BG_IMG = "bg_light.png"
L_BG_URL = f"{IMAGE_URL}bg_light.png"

LOGO = "logo.png"
LOGO_URL = f"{IMAGE_URL}logo.png"
TARGET = "target.png"
TARGET_PATH = path.join(ASSET_IMAGE_DIR, TARGET)
TARGET_URL = f"{IMAGE_URL}target.png"
BAR_IMG = "bar.png"
BAR_URL = f"{IMAGE_URL}bar.png"

BGM_PATH = path.join(MUSIC_DIR, "BGM.mp3")
BGM_URL = MUSIC_URL + "BGM.mp3"
BOMB_PATH = path.join(SOUND_DIR, "bomb.mp3")
BOMB_URL = SOUND_URL + "bomb.mp3"
HELP_TXT_STYLE = "16px Arial BOLD"
HELP_TXT_COLOR = "#43A047"



from os import path

import pygame.mixer


def create_sounds_data(id: str, name: str):
    return {
        "_id": id
        , "_name": name
    }


def create_bgm_data(name: str, volume: float):
    return {
        "_name": name
        , "_volume": volume
    }


class SoundController:
    def __init__(self, sound_path: str, sounds_data_list: list):
        self._sound_path = sound_path
        if not self._sound_path:
            return
        self._sounds_obj = {}
        pygame.mixer.init()
        for sounds_data in sounds_data_list:
            sound_data = path.join(self._sound_path, sounds_data["_name"])
            self._sounds_obj[sounds_data["_id"]] = pygame.mixer.Sound(sound_data)

    def play_music(self, bgm_data: dict) -> None:
        if not self._sound_path:
            return
        pygame.mixer.init()
        pygame.mixer.music.load(path.join(self._sound_path, bgm_data["_name"]))
        pygame.mixer.music.set_volume(bgm_data["_volume"])
        pygame.mixer.music.play(-1)

    def play_sound(self, id: str, volume: float, maz_time: int) -> None:
        if not self._sound_path:
            return
        sound_obj = self._sounds_obj[id]
        sound_obj.set_volume(volume)
        sound_obj.play(maxtime=maz_time)



import sys
from functools import cache

import loguru

_logger = None


@cache
def get_singleton_logger():
    global _logger
    if _logger is None:
        # logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.DEBUG)

        _logger = loguru.logger

        _logger.remove()
        _logger.add(
            sys.stdout,
            level="INFO",
            format='<green>{time:YYYY-MM-DD HH:mm:ss.SSSSS}</green> | <level>{level: <8}</level> | <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>'
        )


    return _logger


logger = get_singleton_logger()



import json
from mlgame.core.communication import TransitionCommManager
from mlgame.core.model import GameProgressSchema, MLGameDataType
from mlgame.executor.interface import ExecutorInterface
from mlgame.utils.logger import logger
import os


class ProgressLogExecutor(ExecutorInterface):
    def __init__(self, progress_folder, progress_frame_frequency, pl_comm: TransitionCommManager):
        # super().__init__(name="ws")
        self._proc_name = f"progress_log({progress_folder}"
        self._progress_folder = progress_folder
        self._progress_frame_frequency = progress_frame_frequency
        self._comm_manager = pl_comm
        self._recv_data_func = self._comm_manager.recv_from_game
        self._filename = "{}.json"
        self._progress_data = []


    def save_json_and_init(self, path):
        with open(path, 'w') as f:
            data = json.dumps(self._progress_data, default=lambda o: o.model_dump() if hasattr(o, "model_dump") else o)
            # json.dump(self._progress_data, f)
            f.write(data)
        # Get the file size in kilobytes (1 KB = 1024 bytes)
        file_size_kb = os.path.getsize(path) / 1024
        # Print the file path and file size in KB
        logger.info(f"File saved to: {path}, file size: {file_size_kb:.2f} KB")

        self._progress_data = []

    def run(self):
        self._comm_manager.start_recv_obj_thread()

        try:
            progress_count = -1
            # Process data until we receive the game result
            while True:
                game_data = self._recv_data_func()
                # logger.info(f"game_data_type: {game_data.type}")
                
                if game_data.type == MLGameDataType.GAME_RESULT:
                    break
                    
                if game_data.type == MLGameDataType.GAME_PROGRESS:
                    
                    progress_data = GameProgressSchema.model_validate(game_data.data)
                    frame = progress_data.frame
                    # Check if we need to save the progress data
                    if (frame - 1) % self._progress_frame_frequency == 0 and frame != 1:
                        progress_count += 1
                        filepath = os.path.join(
                            self._progress_folder, 
                            self._filename.format(progress_count)
                        )
                        self.save_json_and_init(filepath)
                        
                    self._progress_data.append(progress_data)
            
            # Handle remaining progress data after the game is finished
            if self._progress_data:
                progress_count += 1
                filename = self._filename.format(str(progress_count) + '-end')
                filepath = os.path.join(self._progress_folder, filename)
                self.save_json_and_init(filepath)
                
        except Exception as e:
            # exception = TransitionProcessError(self._proc_name, traceback.format_exc())
            self._comm_manager.send_exception(
                f"exception on {self._proc_name}")
            # catch connection error
            logger.exception(e)
        finally:
            logger.info("end pl")



import pygame
from loguru import logger

from .playingMode import PlayingMode
from .coinPlayMode import CoinMode
from .reliveMode import ReliveMode
from mlgame.view.decorator import check_game_progress, check_game_result
from mlgame.view.view_model import create_text_view_data, create_asset_init_data, create_image_view_data, \
    Scene, create_rect_view_data
from mlgame.game.paia_game import PaiaGame

from .env import *
from .sound_controller import *

'''need some fuction same as arkanoid which without dash in the name of fuction'''


class RacingCar(PaiaGame):
    def __init__(self, user_num: int, game_type, car_num, racetrack_length=15000, rounds=1, sound="off",*args,**kwargs):
        super().__init__(user_num=user_num)
        self.game_times_goal = rounds
        self.game_times = 1
        self.score = []  # 用於計算積分
        self.is_sound = sound
        self.cars_num = car_num
        self.sound_controller = SoundController(self.is_sound)
        self.game_type = game_type
        self.user_num = user_num
        self.length = racetrack_length
        self.game_mode = self.set_game_mode()
        self.game_mode.sound_controller.play_music()
        self.scene = Scene(WIDTH, HEIGHT, BLACK)
        self.attachements = []

    def get_data_from_game_to_player(self) -> dict:
        scene_info = self.get_scene_info
        to_player_data = {}
        for user in self.game_mode.users:
            player_data = user.get_info()
            player_data["all_cars_pos"] = scene_info["cars_pos"]
            player_data["frame"] = scene_info["frame"]
            # player_data["status"] = scene_info["status"]
            if self.game_type == "COIN":
                player_data["coin"] = scene_info["coin"]
            else:
                player_data["coin"] = []
            to_player_data[str(player_data["id"] + 1) + "P"] = player_data

        if to_player_data:
            return to_player_data
        else:
            return {
                "1P": scene_info,
                "2P": scene_info,
                "3P": scene_info,
                "4P": scene_info
            }

    @property
    def get_scene_info(self):
        """
        Get the scene information
        """
        cars_pos = []
        computer_cars_pos = []

        scene_info = {
            "frame": self.game_mode.frame,
            "status": self.game_mode.state,
            "background": [(self.game_mode.bg_x, 0), (self.game_mode.rel_x, 0)], }

        for user in self.game_mode.cars:
            car_info = user.get_info()
            cars_pos.append((car_info["x"], car_info["y"]))
            if car_info["id"] <= 4:
                scene_info["player_" + str(car_info["id"]) + "_pos"] = (car_info["x"], car_info["y"])
            elif car_info["id"] > 100:
                computer_cars_pos.append((car_info["x"], car_info["y"]))
        scene_info["computer_cars"] = computer_cars_pos
        scene_info["cars_pos"] = cars_pos

        if self.game_type == "COIN":
            coin_pos = []
            for coin in self.game_mode.coins:
                coin_pos.append(coin.get_position())
            scene_info["coin"] = coin_pos

        scene_info["game_result"] = self.game_mode.winner
        return scene_info

    def update(self, commands):
        self.frame_count += 1
        self.game_mode.handle_event()
        self.game_mode.detect_collision()
        self.game_mode.update(commands)
        self.game_result_state = self.game_mode.state
        if not self.isRunning():
            # collect game rank
            game_result = self.game_mode.winner.copy()
            if len(self.attachements) == 0:
                """ fisrt time end"""
                self.attachements = game_result
                for user in self.attachements:
                    user["accumulated_score"] = 5 - user["single_rank"]

            else:
                for user in self.attachements:
                    for single_rank in game_result:
                        if single_rank['player_num'] == user['player_num']:
                            match_single_rank = single_rank
                    user["accumulated_score"] += (5 - match_single_rank["single_rank"])
                    user["single_rank"] = match_single_rank["single_rank"]
                #  TODO sort by accumulated_score and add rank in each user
                # Sort users by accumulated_score in descending order and add rank
            sorted_users = sorted(self.attachements, key=lambda x: x['accumulated_score'], reverse=True)

            for rank, user in enumerate(sorted_users, start=1):
                user["rank"] = rank

            if self.game_times < self.game_times_goal:
                self.game_times += 1
                # next run
                self.frame_count = 0
                self.game_mode = self.set_game_mode()
                self.game_mode.sound_controller.play_music()

                # self.reset()
                # return "RESET"
            else:
                return "RESET"

    def reset(self):
        self.frame_count = 0
        self.game_mode = self.set_game_mode()
        self.attachements = []
        self.game_mode.sound_controller.play_music()
        logger.info("Start new run")

    def isRunning(self):
        return self.game_mode.isRunning()

    def get_scene_init_data(self) -> dict:
        """
        Get the scene and object information for drawing on the web
        """
        game_info = {"scene": self.scene.__dict__,
                     "assets": []}
        sys_car_path = path.join(ASSET_IMAGE_DIR, COMPUTER_CAR_IMAGE[0])
        game_info["assets"].append(
            create_asset_init_data("computer_car", car_size[0], coin_size[1], sys_car_path, COMPUTER_CAR_URL))
        for i in range(self.user_num):
            # TODO bad car
            game_info["assets"].append(
                create_asset_init_data("player" + str(i + 1) + "_car", car_size[0], coin_size[1],
                                       path.join(ASSET_IMAGE_DIR, USER_IMAGE[i][0]), USER_CAR_URL[i]))
        game_info["assets"].append(
            create_asset_init_data("background", 2000, HEIGHT, path.join(ASSET_IMAGE_DIR, BACKGROUND_IMAGE[0]),
                                   BACKGROUND_URL))
        game_info["assets"].append(
            create_asset_init_data("start_line", 120, 450, path.join(ASSET_IMAGE_DIR, START_LINE_IMAGE[0]), START_URL))
        game_info["assets"].append(
            create_asset_init_data("finish_line", 120, 450, path.join(ASSET_IMAGE_DIR, START_LINE_IMAGE[1]), FINISH_URL))
        if self.game_type == "COIN":
            game_info["assets"].append(
                create_asset_init_data("coin", coin_size[0], coin_size[1], path.join(ASSET_IMAGE_DIR, COIN_IMAGE),
                                       COIN_URL))
            game_info["assets"].append(
                create_asset_init_data("info_coin", 513, 80, path.join(ASSET_IMAGE_DIR, RANKING_IMAGE[0]),
                                       INFO_COIN_URL))
        else:
            game_info["assets"].append(
                create_asset_init_data("info_km", 513, 80, path.join(ASSET_IMAGE_DIR, RANKING_IMAGE[1]), INFO_KM_URL))

        return game_info

    @check_game_progress
    def get_scene_progress_data(self) -> dict:
        """
        Get the position of src objects for drawing on the web
        """
        scene_info = self.get_scene_info
        game_progress = {
            "frame": self.frame_count,
            "background": [],
            "object_list": [],
            "toggle_with_bias": [],
            "toggle": [],
            "foreground": [],
            "user_info": [],
            "game_sys_info": {}
        }
        if self.game_type == "COIN":
            info_board = create_image_view_data("info_coin", WIDTH - 517, 5, 513, 80)
        else:
            info_board = create_image_view_data("info_km", WIDTH - 517, 5, 513, 80)
        game_progress["foreground"].append(info_board)
        bg1 = create_image_view_data("background", self.game_mode.bg_x, 0, 2000, HEIGHT)
        bg2 = create_image_view_data("background", self.game_mode.rel_x, 0, 2000, HEIGHT)
        game_progress["object_list"].append(bg1)

        if self.game_mode.rel_x <= WIDTH:
            game_progress["object_list"].append(bg2)
        # 縮圖
        block = create_rect_view_data("block", 0, 650, 1000, 50, BLACK)
        game_progress["foreground"].append(block)
        # line and lane
        game_progress["object_list"].append(self.game_mode.line.get_asset_info())
        # for lane in self.game_mode.lanes:
        #     game_progress["object_list"].append(lane.get_asset_info())
        # user
        for user in self.game_mode.users:
            user_image = create_image_view_data("player" + str(user.car_no + 1) + "_car", user.rect[0], user.rect[1],
                                                car_size[0], car_size[1])
            game_progress["object_list"].append(user_image)
            point = create_rect_view_data("user", round(user.distance * (900 / self.length)),
                                          650 + round(user.rect.top * (50 / 500)),
                                          4, 4, USER_COLOR[user.car_no])
            game_progress["foreground"].append(point)
        # score
        if self.game_type == "COIN":
            for coin in scene_info["coin"]:
                coin_image = create_image_view_data("coin", coin[0], coin[1], coin_size[0], coin_size[1])
                game_progress["object_list"].append(coin_image)
            for ind,user in enumerate(self.game_mode.users,start=1):
                score = create_text_view_data(f"P{ind}   {round(user.coin_num):02d}", 560 + user.car_no * 110, 20, WHITE, "16px Arial")
                game_progress["foreground"].append(score)
        else:
            for ind,user in enumerate(self.game_mode.users,start=1):
                score = create_text_view_data( f"P{ind}  {round(user.distance):5d}m", 560 + user.car_no * 110, 20, WHITE,
                                              "16px Arial")
                game_progress["foreground"].append(score)
        # computer car
        for car in scene_info["computer_cars"]:
            car_image = create_image_view_data("computer_car", car[0], car[1], car_size[0], car_size[1])
            game_progress["object_list"].append(car_image)
        return game_progress

    @check_game_result
    def get_game_result(self):
        """
        Get the src result for the web
        """

        return {"frame_used": self.frame_count,
                "status": self.game_result_state,
                "attachment": self.rank()
                }

    def get_keyboard_command(self):
        """
        Get the command according to the pressed keys
        """
        key_pressed_list = pygame.key.get_pressed()
        cmd_1P = []
        cmd_2P = []
        cmd_3P = []
        cmd_4P = []

        if key_pressed_list[pygame.K_LEFT]: cmd_1P.append(BRAKE_cmd)
        if key_pressed_list[pygame.K_RIGHT]: cmd_1P.append(SPEED_cmd)
        if key_pressed_list[pygame.K_UP]: cmd_1P.append(LEFT_cmd)
        if key_pressed_list[pygame.K_DOWN]: cmd_1P.append(RIGHT_cmd)

        if key_pressed_list[pygame.K_a]: cmd_2P.append(BRAKE_cmd)
        if key_pressed_list[pygame.K_d]: cmd_2P.append(SPEED_cmd)
        if key_pressed_list[pygame.K_w]: cmd_2P.append(LEFT_cmd)
        if key_pressed_list[pygame.K_s]: cmd_2P.append(RIGHT_cmd)

        if not self.isRunning():
            return {"1P": "RESET",
                    "2P": "RESET",
                    "3P": "RESET",
                    "4P": "RESET",
                    }

        return {"1P": cmd_1P,
                "2P": cmd_2P,
                "3P": cmd_3P,
                "4P": cmd_4P,
                }

    @staticmethod
    def ai_clients():
        """
        let MLGame know how to parse your ai,
        you can also use this names to get different cmd and send different data to each ai client
        """
        return [
            {"name": "1P"},
            {"name": "2P"},
            {"name": "3P"},
            {"name": "4P"}
        ]

    def set_game_mode(self):
        if self.game_type == "NORMAL":
            game_mode = PlayingMode(self.user_num, self.cars_num, self.length, self.sound_controller)
        elif self.game_type == "COIN":
            game_mode = CoinMode(self.user_num, self.cars_num, self.length, self.sound_controller)
        elif self.game_type == "RELIVE":
            game_mode = ReliveMode(self.user_num, self.cars_num, self.length, self.sound_controller)

        return game_mode

    def rank(self):
        # game_result = self.get_scene_info["game_result"]
        # TODO refactor
        # if len(self.attachements)==0:
        #     self.attachements = game_result
        #     for user in self.attachements:
        #         user["accumulated_score"] = 5 - user["rank"]
        #     return self.attachements
        #
        # for user in self.attachements:
        #     for single_rank in game_result:
        #         if single_rank['player'] == user['player']:
        #             match_single_rank = single_rank
        #     user["accumulated_score"] += (5 - match_single_rank["rank"])
        # if self.score:
        #     for user in self.score:
        #         user["accumulated_score"] += (5 - user["rank"])
        # else:
        #     self.score = single_game_result
        #     for user in self.score:
        #         user["accumulated_score"] = 5 - user["rank"]

        return self.attachements



from os import path
import random
import math


PLAYER_NUM = 1
OIL_NUM = 2
BULLET_NUM = 2
MAP_DIR = path.join(path.dirname(__file__), "..", "asset", 'maps')
MAP_VERSION = "1.9"
TILED_VERSION = "1.9.2"


class MapGenerator:

    def __init__(self, green_team_num : int, blue_team_num : int, width : int, height : int) -> None:

        self.green_team_num = green_team_num
        self.blue_team_num = blue_team_num

        # height should be a factor of 600
        # default is 20, 12

        self.width = width
        self.height = height

        self.height_per_tile = math.floor(600/self.height)
        self.width_per_tile = self.height_per_tile

        self.screen_width = self.width * self.height_per_tile
        self.screen_height = 700

    def getTileSize(self) -> int:
        return self.height_per_tile
    
    def getScreeenSize(self) -> tuple:
        return self.screen_width, self.screen_height

    def pos2index(self, x : int, y : int) -> int:
        return y * (self.width * 2 + 1) + x * 2

    def mirrored_pos(self, x : int, y : int) -> tuple:
        return self.width - x - 1, self.height - y - 1

    def random_pos(self, map_arr) -> tuple:
        random_x, random_y = random.randint(1, self.width-2), random.randint(1, self.height-2)
        mir_x, mir_y = self.mirrored_pos(random_x, random_y)
        while map_arr[random_y][random_x] != 0 or map_arr[mir_y][mir_x] != 0:
            random_x, random_y = random.randint(1, self.width-2), random.randint(1, self.height-2)
            mir_x, mir_y = self.mirrored_pos(random_x, random_y)

        return random_x, random_y

    def generate_map_str(self) -> str:

        # generate default map[y][x]
        map_arr = [[0 for _ in range(self.width)] for _ in range(self.height)]
        
        for x in range(self.width):
            map_arr[0][x] = 3
            map_arr[self.height-1][x] = 3
        
        for y in range(self.height):
            map_arr[y][0] = 3
            map_arr[y][self.width-1] = 3

        if self.width % 2 == 1:
            for y in range(1, self.height-1):
                map_arr[y][self.width//2] = 3
        else:
            for y in range(1, self.height):
                if y < math.ceil(self.height/2):
                    map_arr[y][self.width//2-1] = 3
                else:
                    map_arr[y][self.width//2] = 3

        # add player
        for _ in range(min(self.green_team_num, self.blue_team_num)):
            rand_x, rand_y = self.random_pos(map_arr)
            while rand_x < self.width // 2:
                rand_x, rand_y = self.random_pos(map_arr)
            mir_x, mir_y = self.mirrored_pos(rand_x, rand_y)

            map_arr[rand_y][rand_x] = 1
            map_arr[mir_y][mir_x] = 2

        # add remaining green team
        for _ in range(self.green_team_num - self.blue_team_num):
            rand_x, rand_y = self.random_pos(map_arr)
            while rand_x < self.width // 2:
                rand_x, rand_y = self.random_pos(map_arr)

            map_arr[rand_y][rand_x] = 1

        # add remaining blue team
        for _ in range(self.blue_team_num - self.green_team_num):
            rand_x, rand_y = self.random_pos(map_arr)
            while rand_x >= self.width // 2:
                rand_x, rand_y = self.random_pos(map_arr)

            map_arr[rand_y][rand_x] = 2
        
        # add bullet station
        for _ in range(BULLET_NUM):
            rand_x, rand_y = self.random_pos(map_arr)
            mir_x, mir_y = self.mirrored_pos(rand_x, rand_y)
            map_arr[rand_y][rand_x] = 4
            map_arr[mir_y][mir_x] = 4

        # add oil station
        for _ in range(OIL_NUM):
            rand_x, rand_y = self.random_pos(map_arr)
            mir_x, mir_y = self.mirrored_pos(rand_x, rand_y)
            map_arr[rand_y][rand_x] = 5
            map_arr[mir_y][mir_x] = 5

        map_str = ""
        
        for row in map_arr:
            for id in row:
                map_str += str(id) + ","
            map_str += "\n"
        map_str = map_str[:-2]

        return map_str
    
    def generate_map(self):
        map_name = f"map_{self.green_team_num}_v_{self.blue_team_num}.tmx"
        map_path = path.join(MAP_DIR, map_name)
        print(f'generate map at : {map_path}', flush=True)

        with open(map_path, "w") as file:
            # file.write("test")
            file.write(f"""\
<?xml version="1.0" encoding="UTF-8"?>
<map version="{MAP_VERSION}" tiledversion="{TILED_VERSION}" orientation="orthogonal" renderorder="right-down" width="{self.width}" height="{self.height}" tilewidth="{self.width_per_tile}" tileheight="{self.height_per_tile}" infinite="0" nextlayerid="2" nextobjectid="1">
 <tileset firstgid="1" name="TankManObj" tilewidth="{self.width_per_tile}" tileheight="{self.height_per_tile}" tilecount="5" columns="5">
  <image source="../image/TankManObj.png" width="250" height="50"/>
 </tileset>
 <layer id="1" name="layer 1" width="{self.width}" height="{self.height}">
  <data encoding="csv">
{self.generate_map_str()}
  </data>
 </layer>
</map>
                """)


if __name__ == "__main__":
    map_generator = MapGenerator(1, 1)
    map_generator.generate_map()
    




import abc

import pygame

class I_Commander(abc.ABC):
    @abc.abstractmethod
    def getControlDict(self) -> dict:
        pass

keyboardSet = [
    {"MOVE_LEFT": pygame.K_UP,
     "MOVE_RIGHT": pygame.K_DOWN,
     "SPEED": pygame.K_RIGHT,
     "BRAKE": pygame.K_LEFT},

    {"MOVE_LEFT": pygame.K_w,
     "MOVE_RIGHT": pygame.K_s,
     "SPEED": pygame.K_d,
     "BRAKE": pygame.K_a},


    {"MOVE_LEFT": pygame.K_SPACE,
     "MOVE_RIGHT": pygame.K_SPACE,
     "SPEED": pygame.K_SPACE,
     "BRAKE": pygame.K_SPACE},

    {"MOVE_LEFT": pygame.K_SPACE,
     "MOVE_RIGHT": pygame.K_SPACE,
     "SPEED": pygame.K_SPACE,
     "BRAKE": pygame.K_SPACE}

]

class KeyBoardCommander(I_Commander):
    def __init__(self, keyboard_no=0):
        self.no = keyboard_no
        self.speedKey = keyboardSet[keyboard_no]["SPEED"]
        self.brakeKey = keyboardSet[keyboard_no]["BRAKE"]
        self.moveLeftKey = keyboardSet[keyboard_no]["MOVE_LEFT"]
        self.moveRightKey = keyboardSet[keyboard_no]["MOVE_RIGHT"]

    def getControlDict(self):
        keys = pygame.key.get_pressed()
        control_list = []
        control_dic = {"LEFT": keys[self.moveLeftKey],
                       "RIGHT": keys[self.moveRightKey],
                       "SPEED_UP": keys[self.speedKey],
                       "BRAKEDOWN": keys[self.brakeKey]}
        if control_dic["LEFT"]:
            control_list.append("MOVE_LEFT")
        if control_dic["RIGHT"]:
            control_list.append("MOVE_RIGHT")
        if control_dic["SPEED_UP"]:
            control_list.append("SPEED")
        if control_dic["BRAKEDOWN"]:
            control_list.append("BRAKE")

        return control_list




# Uncatched python exception
UNHANDLED_ERROR = 1
# An error occurred while handling the execution command
COMMAND_LINE_ERROR = 2
# An error occurred while executing the game
GAME_EXECUTION_ERROR = 3
# An error occurred while compiling the non-python script
COMPILATION_ERROR = 4

ML_PROCESS_ERROR=100


"""
The template of the main script of the machine learning process
"""

import pygame
from ml.test import *

class MLPlay:
    def __init__(self, ai_name, *args, **kwargs):
        """
        Constructor

        @param ai_name A string "1P" or "2P" indicates that the `MLPlay` is used by
               which side.
        """
        self.side = ai_name
        print(f"Initial Game {ai_name} ml script")
        self.time = 0

    def update(self, scene_info: dict, keyboard=[], *args, **kwargs):
        """
        Generate the command according to the received scene information
        """
        # if scene_info["used_frame"] == 1:
        #     print(f"{self.side}_scene_info_{scene_info.keys()}")
        #     if scene_info["id"] == "1P":
        #         print(f'1P_competitor_info{scene_info["competitor_info"]}')
        #     else:
        #         print(f'2P_competitor_info{scene_info["competitor_info"]}')
        # print(keyboard)
        if scene_info["status"] != "GAME_ALIVE":
            # print(scene_info)
            return "RESET"

        command = []
        if self.side == "1P":
            if pygame.K_RIGHT in keyboard:
                command.append("TURN_RIGHT")
            elif pygame.K_LEFT in keyboard:
                command.append("TURN_LEFT")
            elif pygame.K_UP in keyboard:
                command.append("FORWARD")
            elif pygame.K_DOWN in keyboard:
                command.append("BACKWARD")

            if pygame.K_z in keyboard:
                command.append("AIM_LEFT")
            elif pygame.K_x in keyboard:
                command.append("AIM_RIGHT")

            if pygame.K_m in keyboard:
                command.append("SHOOT")
            # debug
            if pygame.K_b in keyboard:
                command.append("DEBUG")
            # paused
            if pygame.K_t in keyboard:
                command.append("PAUSED")
        elif self.side == "2P" or self.side == "4P":
            if pygame.K_d in keyboard:
                command.append("TURN_RIGHT")
            elif pygame.K_a in keyboard:
                command.append("TURN_LEFT")
            elif pygame.K_w in keyboard:
                command.append("FORWARD")
            elif pygame.K_s in keyboard:
                command.append("BACKWARD")

            if pygame.K_q in keyboard:
                command.append("AIM_LEFT")
            elif pygame.K_e in keyboard:
                command.append("AIM_RIGHT")

            if pygame.K_f in keyboard:
                command.append("SHOOT")
        bullet = Bullet()
        # print(scene_info["gun_angle"])
        # print(bullet.is_wall_in_bullet_range({"x": scene_info["x"], "y": scene_info["y"]}, scene_info["gun_angle"], scene_info["walls_info"], 200))
        if not command:
            command.append("NONE")
        
        return command

    def reset(self):
        """
        Reset the status
        """
        print(f"reset Game {self.side}")



from pathlib import Path
from typing import Literal

import pydantic
from pydantic import AnyHttpUrl, BaseModel, Field


class MusicInitSchema(BaseModel):
    type: Literal["music"] = Field(default="music")
    music_id: str
    file_path: Path  # This will ensure the path is valid
    url: AnyHttpUrl



class SoundInitSchema(BaseModel):
    type: Literal["sound"] = Field(default="sound")
    sound_id: str
    file_path: Path  # This will ensure the path is valid
    url: AnyHttpUrl




class SoundProgressSchema(pydantic.BaseModel):
    type: Literal["sound"] = Field(default="sound")
    sound_id: str




class MusicProgressSchema(pydantic.BaseModel):
    type: Literal["music"] = Field(default="music")  # Default and required
    music_id: str




def create_music_init_data(music_id: str, file_path: str, github_raw_url: str):
    # assert file_path is valid
    return {
        "type": "music",
        "music_id": music_id,
        "file_path": file_path,
        "url": github_raw_url,
    }


def create_sound_init_data(sound_id: str, file_path: str, github_raw_url: str):
    # assert file_path is valid
    return {
        "type": "sound",
        "sound_id": sound_id,
        "file_path": file_path,
        "url": github_raw_url,
    }



"""
The template of the main script of the machine learning process
"""


class MLPlay:
    def __init__(self,ai_name, *args, **kwargs):
        """
        Constructor
        """
        print(ai_name)

    def update(self, scene_info, *args, **kwargs):
        """
        Generate the command according to the received `scene_info`.
        """
        # Make the caller to invoke `reset()` for the next round.
        if (scene_info["status"] == "GAME_OVER" or
                scene_info["status"] == "GAME_PASS"):
            return "RESET"
        if not scene_info["ball_served"]:
            command = "SERVE_TO_LEFT"
        else:
            command = "MOVE_LEFT"

        return command

    def reset(self):
        """
        Reset the status
        """
        self.ball_served = False



import pygame
from src import MazeCar

from src.env import FPS
from mlgame.view.view import PygameView
from mlgame.game.generic import quit_or_esc

SOUND_OFF = "off"

MOVE_MAZE = "MOVE_MAZE"

if __name__ == '__main__':
    pygame.init()
    # game = MazeCar.MazeCar(6, "MAZE", 6, 10000, 3, "off")
    # game = MazeCar.MazeCar(user_num=2, game_type=MOVE_MAZE, map=5, time_to_play=2000, sensor_num=3, sound=SOUND_OFF)
    game = MazeCar.MazeCar(2, "PRACTICE", 12, 10000, 5, "off")
    scene_init_info_dict = game.get_scene_init_data()
    game_view = PygameView(scene_init_info_dict)
    frame_count = 0
    while game.is_running and not quit_or_esc():
        pygame.time.Clock().tick_busy_loop(FPS)
        game.update(game.get_keyboard_command())
        game_progress_data = game.get_scene_progress_data()
        game_view.draw(game_progress_data)
        frame_count += 1

    pygame.quit()



import numpy as np

"""
update：20220718
"""
class autoRCar:
    def __init__(self):
        #本車編號
        self.player_id = 0
        #本車目前賽道和位置
        self.myLane = 1
        self.myDist = 0
        self.myCar = 0 #絕對位置
        self.myPos = 0 #相對位置 (一般pos=11)
        #本車目標賽道
        self.targetLane = 0
        #其他車資訊
        self.ROW = 9    #賽道列
        self.COL = 0   #賽道行 
        #常數
        self.CAR_HEIGHT = 60 #車長
        self.CAR_WIDTH = 30 #車寬30
        self.LANE_LENGTH = 80  #賽道長 
        self.LANE_WIDTH = 50  #賽道寬
        self.LANE_OFFSET = 20 #判斷車道號碼寬容度
        self.DIST_OFFSET = 15 #判斷車道號碼寬容度

        self.SAFE = 0 
        self.PCCAR = int(1E2)
        self.UNDEF = int(1E1)
        self.COIN  = -int(1E1)
        self.MYCAR  = -int(1E2)

        self.LEFT = -self.COL 
        self.RIGHT = self.COL 
        self.UP = 1 
        self.DOWN = -1 
        #dir list
        self.KEY_DIRC = []
        self.COIN_DIRC =[]
        self.PC_DIRC =[]
        
        self.pcCar =[]
        self.coins=[]
        # self.notBestL=[] 
        self.board =[]  # 一維陣列
        #收集資料
        self.feature =[]
        self.target=[]
        # fname = time.strftime("%Y-%m-%d_%H-%M-%S") + ".txt"
        # file_newname = path.join(path.dirname(__file__), fname)
        # self.f = open(file_newname, "w")
        self.count = 1

    #block function 1
    def getCarInfo(self, scene_info, lane_size=60 ,feature_size=5):  # , LANE_LENGTH, COL
        """
        取得電腦車和其他玩家位置資訊和賽道
        """
        # if feature_size < 3 or feature_size > 9  or lane_size < 60 or lane_size >=100:
        #     # print("wrong parameters：feature_size(3-9)，lane_size(60-120)");
        #     return False,"參數錯誤：feature_size=(3~9)，lane_size=(60~120)"
        #更新常數
        self.COL = feature_size
        self.LANE_LENGTH = lane_size
        # 定義方向
        self.LEFT = -self.COL
        self.RIGHT = self.COL
        self.KEY_DIRC = [self.DOWN,self.UP, self.LEFT, self.RIGHT]
        self.COIN_DIRC =[self.DOWN, self.LEFT, self.RIGHT]
        self.PC_DIRC =[self.DOWN, self.UP]
        #確定特徵大小
        self.board = [0] * self.COL * self.COL 
        self.feature=[]
        self.player_id = scene_info['id']
        self.x = scene_info['x']
        self.y = scene_info['y']
        # 取得本車的資訊velocity：速度，coin_num：金幣數，distance：累積距離
        # car_vel = scene_info["velocity"]
        # car_dis = scene_info["distance"]
        # coin_num = scene_info["coin_num"]
        
        # 金幣位置
        self.coins.clear()
        if scene_info["coin"]:
            # print(scene_info["coin"])
            for i, coin in enumerate(scene_info["coin"]):
                if coin[0] > self.x-self.LANE_LENGTH and coin[0] <= self.x+(self.COL-1)*self.LANE_LENGTH-self.DIST_OFFSET:
                    lane = self.getLane(coin[1], 5)  # y
                    dist = self.getDist(coin[0])  # x
                    # if dist == -1:
                    #     print("coin dist wrong:", x, coin[0])
                    # else:
                    self.coins.append(lane * self.COL + dist)
        
        self.pcCar.clear()
        for i, car in enumerate(scene_info['all_cars_pos']):
            # 取得本車的資訊
            if i == self.player_id:
                self.myLane = self.getLane(self.y, self.player_id)
                self.myDist = 1  
                self.startLane = self.getStartLane(self.myLane)
                self.myCar = self.myLane * self.COL + self.myDist  # 取得1D座標
                self.myPos = self.myCar - self.startLane*self.COL
            else: 
                # 其他車範圍
                if car[0] > self.x-self.LANE_LENGTH-self.DIST_OFFSET and car[0] <= self.x+(self.COL-1)*self.LANE_LENGTH-self.DIST_OFFSET:
                    lane = self.getLane(car[1], 5)  # y
                    dist = self.getDist(car[0])  # x
                    # if dist == -1 or lane == -1:
                    #     print("range:",x-LANE_LENGTH-DIST_OFFSET,x+(COL-1)*LANE_LENGTH-DIST_OFFSET)
                    #     print("lane dist wrong:", x, car)
                    # else:
                    self.pcCar.append(lane * self.COL + dist)
        
        self.initBoard(self.board, self.pcCar, self.coins)
        # print(self.myPos,self.myCar,self.board)
        self.feature = self.transBoard() #self.myLane, self.board
        return self.feature, self.myLane
    
    #block function 2
    def getTarget(self):#best move：return target number 1 -1 COL -COL
        self.findPathBfs(self.board, self.pcCar, self.coins)
        best_mv = self.minMV( self.myPos, self.board)
        cmd,targeL = self.transCmd(best_mv)
        return best_mv, targeL
    
    #block function 3 
    def printBoard(self):
        b = np.array([i for i in range(self.COL)])  # COL
        for i in range(self.COL):
            a = np.array(self.feature[i*self.COL: i*self.COL+self.COL])
            b = np.vstack((b, a))
        b =  np.delete(b, 0, 0) #刪除第一行
        print(b)
        
    #block function 4 位置是否已達到賽道中心位  input：車的y, 賽道 (本車)
    def isCenterLane(self,lane_num: int): #, car_y: int, lane_num: int):
        # 計算賽道的中心位，第0道100
        lane = 100+lane_num*self.LANE_WIDTH
        if lane <= self.y and self.y <= lane+self.LANE_OFFSET:
            return True
        else:
            return False
        
        
    def initBoard(self,__board:list, __pcCar:list, __coins:list):
        inRange = self.startLane*self.COL
        for i in range(self.COL * self.COL):
            if i+inRange in __pcCar:
                __board[i] = self.PCCAR  
            elif i+inRange in __coins:
                __board[i] = self.COIN 
            else:
                __board[i] = self.UNDEF 
        __board[self.myPos] = self.MYCAR 
        
    def getStartLane(self, __myLane): #Lane：0-8
        if self.COL==9:
            row_start =0
        else:
            row_start = __myLane - self.COL//2
            while True:
                if  row_start < 0 : 
                    row_start += 1
                elif row_start + self.COL > 9:
                    row_start-=1
                else:
                    break
        return row_start
    
    def transBoard(self): #__myLane,__board:list
        feature = []
        for i in range(0,self.COL*self.COL):
            if self.board[i]==self.UNDEF:
                feature.append(self.SAFE)
            else:
                feature.append(self.board[i])
        return feature


    def findPathBfs(self, __board:list, __pcCar:list,__coins: list):
        inRange = self.startLane*self.COL
        for i in range(self.COL* self.COL):
            if __board[i] == self.UNDEF:
                if i//self.COL == 0 or i//self.COL == self.COL-1:  # 上下兩邊車道penalty較高
                    __board[i] = self.COL - (i % self.COL)
                else:  # 其他
                    __board[i] = self.COL - (i % self.COL) - 1
        #處理金幣
        if __coins:
            for i in __coins:
                pos = i-inRange
                if 0 <= pos and pos < self.COL*self.COL :
                    for dirc in self.COIN_DIRC:
                        next_dir = dirc # 金幣三個方向
                        if self.canMove(pos, next_dir) and self.canMove(pos, next_dir-1) and (pos+next_dir-1)!=self.myCar-1 and __board[pos+next_dir] < self.PCCAR :
                            if dirc == self.DOWN:
                                __board[pos+next_dir] -= 2
                            else: 
                                __board[pos+next_dir-1] -= 3
                                if self.canMove(pos, next_dir+next_dir-2) and (pos+next_dir+next_dir-2)!=self.myCar-1: 
                                    __board[pos+next_dir+next_dir-2] -= 3
                            next_dir += dirc
        #設計其他車penalty
        for i in __pcCar: 
            pos = i-inRange
            for dirc in self.PC_DIRC:
                next_dir = dirc
                for j in range(3):  # 考慮的步數
                    if self.canMove(pos, next_dir) and __board[pos+next_dir] < self.PCCAR:
                        if j==0:
                            __board[pos+next_dir] = self.PCCAR-1 
                        elif dirc == self.DOWN : # 電腦車後面
                            __board[pos+next_dir] += 2 
                        next_dir += dirc
        #標記本車
        __board[self.myPos] = self.MYCAR
        
        
    def canMove(self, pos:int, dirc:int):
        row = pos // self.COL
        if pos + dirc >= self.COL*self.COL:
            return False
        elif pos+dirc < 0:
            return False
        elif (dirc == self.UP or dirc== self.DOWN) and (pos+dirc) // self.COL != row :  # 左右邊界判斷         
            return False
        elif dirc == self.LEFT and (pos + dirc)//self.COL < 0 :   # 上邊界判斷
            return False
        elif dirc == self.RIGHT and (pos + dirc)//self.COL >= self.COL:  # 下邊界判斷
            return False
        else:
            return True

    def minMV(self, __myPos, __board):
        mini = self.PCCAR*2
        mv = None
        for dirc in self.KEY_DIRC:
            if self.canMove(__myPos, dirc) :
                if  dirc == self.DOWN and __board[__myPos +self.DOWN]*2  < mini:
                    mini = __board[__myPos + self.DOWN]*2 
                    mv = dirc
                elif  dirc == self.UP and self.COL==3 and __board[__myPos + self.UP ]*2 < mini: #如果是3x3的特徵
                    mini = __board[__myPos + self.UP]*2 
                    mv = dirc
                elif __board[__myPos + dirc] + __board[__myPos + dirc + 1 ] < mini:
                    mini = __board[__myPos + dirc] +__board[__myPos + dirc + 1]
                    mv = dirc
        return mv
    
    def transCmd(self, best_move):
        if best_move == self.UP:
            self.targetLane = self.myLane
            return ["SPEED"],self.targetLane
        elif best_move == self.RIGHT:
            self.targetLane = self.myLane+1
            return ["SPEED","MOVE_RIGHT"],self.targetLane
        elif best_move == self.LEFT:
            self.targetLane = self.myLane-1
            return ["SPEED","MOVE_LEFT"],self.targetLane
        elif best_move == self.DOWN:
            self.targetLane = self.myLane
            return ["BREAK"],self.targetLane
    
    
    def getDist(self, car_x:int):
        #位置範圍 劃分COL區:0:0~lane_length,  1:lane_length~lane_length*2 
        for i in range(0,self.COL) :
            if i == 0 and self.isRange(car_x, self.x-self.LANE_LENGTH, -1):
                col = 0
                break
            else:
                if self.isRange(car_x, self.x+(i-1)*self.LANE_LENGTH, -1):  # 車道距離底線
                    col = i
                    break
        else:
            col = -1
        return col


    def getLane(self, car_y:int, id:int): #Lane：0-8
        for i in range(0,9):
            if self.isRange(car_y, 100+i*self.LANE_WIDTH, id): #車道中心判斷
                lane = i
                break
        else:
            if id == self.player_id: #尚未到達指定車道
                lane = self.myLane
            else:
                lane = -1
        return lane


    def isRange(self, pos:int, lane_pos:int, id:int):
        if id == -1:  
            if  lane_pos-self.DIST_OFFSET < pos and pos <= lane_pos+self.LANE_LENGTH-self.DIST_OFFSET:
                return True
            else:
                return False
        else:  # Row車道，範圍較大 100<= pos <= 120
            if  lane_pos < pos and pos <= lane_pos+self.LANE_OFFSET:
                return True
            else:
                return False


    
